# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import warnings

from hestia.list_utils import to_list
from marshmallow import ValidationError, fields, validates_schema

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ops.environments.outputs import OutputsSchema
from polyaxon_schemas.ops.environments.persistence import PersistenceSchema
from polyaxon_schemas.ops.environments.resources import PodResourcesSchema


def validate_configmap_refs(values, is_schema=False):
    if values.get('config_map_refs') and values.get('configmap_refs'):
        raise ValidationError('You should only use `config_map_refs`.')

    if values.get('configmap_refs') and is_schema:
        warnings.warn(
            'The `configmap_refs` parameter is deprecated and will be removed in next release, '
            'please use `config_map_refs` instead.',
            DeprecationWarning)
        values['config_map_refs'] = values.pop('configmap_refs')

    return values


def validate_persistence(values, is_schema=False):
    if values.get('persistence') and (values.get('data_refs') or
                                      values.get('artifact_refs')):
        raise ValidationError('You cannot use `persistence` and  `data_refs` or `artifact_refs`.')

    if values.get('persistence') and is_schema:
        warnings.warn(
            'The `persistence` parameter is deprecated and will be removed in next release, '
            'please use `data_refs` and/or `artifact_refs` instead.',
            DeprecationWarning)
        persistence = values.pop('persistence')
        values['data_refs'] = values.get('data_refs', persistence.data)
        values['artifact_refs'] = to_list(values.get('artifact_refs', persistence.outputs),
                                          check_none=True)
        return values


def validate_outputs(values, is_schema=False):
    if values.get('outputs'):
        warnings.warn(
            'The `outputs` parameter is deprecated and will be removed in next release, '
            'please notice that it will be ignored.',
            DeprecationWarning)
        if is_schema:
            values.pop('outputs')


class EnvironmentSchema(BaseSchema):
    # To indicate which worker/ps index this session config belongs to
    index = fields.Int(allow_none=True)
    resources = fields.Nested(PodResourcesSchema, allow_none=True)
    node_selector = fields.Dict(allow_none=True)
    affinity = fields.Dict(allow_none=True)
    tolerations = fields.List(fields.Dict(), allow_none=True)
    service_account = fields.Str(allow_none=True)
    image_pull_secrets = fields.List(fields.Str(), allow_none=True)
    max_restarts = fields.Int(allow_none=True)
    secret_refs = fields.List(fields.Str(), allow_none=True)
    config_map_refs = fields.List(fields.Str(), allow_none=True)
    configmap_refs = fields.List(fields.Str(), allow_none=True)  # Deprecated
    data_refs = fields.List(fields.Str(), allow_none=True)
    artifact_refs = fields.List(fields.Str(), allow_none=True)
    outputs = fields.Nested(OutputsSchema, allow_none=True)  # Deprecated
    persistence = fields.Nested(PersistenceSchema, allow_none=True)

    @staticmethod
    def schema_config():
        return EnvironmentConfig

    @validates_schema
    def validate_configmap_refs(self, values):
        validate_configmap_refs(values, is_schema=True)

    @validates_schema
    def validate_persistence(self, values):
        validate_persistence(values, is_schema=True)
    #
    # @validates_schema
    # def validate_outputs(self, values):
    #     validate_outputs(values, is_schema=True)


class EnvironmentConfig(BaseConfig):
    """
    Pod environment config.

    Args:
        index: `int | None`. The index of the pod.
        resources: `PodResourcesConfig`.
        node_selector: `dict`.
        affinity: `dict`.
        tolerations: `list(dict)`.
    """
    IDENTIFIER = 'environment'
    SCHEMA = EnvironmentSchema
    REDUCED_ATTRIBUTES = ['index',
                          'resources',
                          'node_selector',
                          'affinity',
                          'tolerations',
                          'service_account',
                          'image_pull_secrets',
                          'max_restarts',
                          'secret_refs',
                          'config_map_refs',
                          'data_refs',
                          'artifact_refs',
                          'outputs']

    def __init__(self,
                 index=None,
                 resources=None,
                 node_selector=None,
                 affinity=None,
                 tolerations=None,
                 service_account=None,
                 image_pull_secrets=None,
                 max_restarts=None,
                 secret_refs=None,
                 config_map_refs=None,
                 configmap_refs=None,
                 data_refs=None,
                 artifact_refs=None,
                 persistence=None,
                 outputs=None,
                 ):
        self.index = index
        self.resources = resources
        self.node_selector = node_selector
        self.affinity = affinity
        self.tolerations = tolerations
        self.service_account = service_account
        self.image_pull_secrets = image_pull_secrets
        self.max_restarts = max_restarts
        self.secret_refs = secret_refs
        validate_configmap_refs({'config_map_refs': config_map_refs,
                                 'configmap_refs': configmap_refs})
        self.config_map_refs = config_map_refs or configmap_refs
        validate_persistence({'persistence': persistence,
                              'data_refs': data_refs,
                              'artifact_refs': artifact_refs})
        self.data_refs = data_refs
        self.artifact_refs = artifact_refs
        # validate_outputs({'outputs': outputs})
        self.outputs = outputs



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema


class ExecutableSchema(BaseSchema):
    execute_at = fields.LocalDateTime(allow_none=True)
    timeout = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return ExecutableConfig


class ExecutableConfig(BaseConfig):
    SCHEMA = ExecutableSchema
    IDENTIFIER = 'executable'
    REDUCED_ATTRIBUTES = ['execute_at', 'timeout']

    def __init__(self, execute_at=None, timeout=None):
        self.execute_at = execute_at
        self.timeout = timeout



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.ops.operators import ForConfig, IfConfig
from polyaxon_schemas.specs import ExperimentSpecification
from polyaxon_schemas.specs.libs.parser import Parser


class TestOperatorConfigs(TestCase):
    def test_for_operator_config(self):
        config_dict = {
            'len': 5,
            'do': 'Value at {{ i }}',
            'index': 'i'
        }
        config = ForConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        expected = ['Value at 0', 'Value at 1', 'Value at 2', 'Value at 3', 'Value at 4']
        assert expected == config.parse(ExperimentSpecification, Parser(), {})

        config_dict = {
            'len': 5,
            'do': [{'Conv2D': {'strides': ['{{ i }}', '{{ i }}']}},
                   {'Pooling2D': {'strides': ['{{ i }}', '{{ i }}']}}],
            'index': 'i'
        }
        config = ForConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        # Lists get flattened
        expected = [{'Conv2D': {'strides': [0, 0]}}, {'Pooling2D': {'strides': [0, 0]}},
                    {'Conv2D': {'strides': [1, 1]}}, {'Pooling2D': {'strides': [1, 1]}},
                    {'Conv2D': {'strides': [2, 2]}}, {'Pooling2D': {'strides': [2, 2]}},
                    {'Conv2D': {'strides': [3, 3]}}, {'Pooling2D': {'strides': [3, 3]}},
                    {'Conv2D': {'strides': [4, 4]}}, {'Pooling2D': {'strides': [4, 4]}}]
        assert expected == config.parse(ExperimentSpecification, Parser(), {})

    def test_if_operator_config(self):
        config_dict = {
            'cond': '{{ i }} == 5',
            'do': 'It was True',
            'else_do': 'It was False'
        }
        config = IfConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert 'It was True' == config.parse(ExperimentSpecification, Parser(), {'i': 5})
        assert 'It was False' == config.parse(ExperimentSpecification, Parser(), {'i': 3})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.logging import LoggingConfig


class TestLoggingConfigs(TestCase):
    def test_logging_config(self):
        config_dict = {
            'level': 'INFO',
            'formatter': None,
        }
        config = LoggingConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import copy


def get_task_values(cluster, is_distributed, values, default_value, task_type):
    result_values = {}
    if not is_distributed:
        return result_values

    result_values = copy.deepcopy(values) if values else {}
    if default_value:
        for i in range(cluster.get(task_type, 0)):
            result_values[i] = values.get(i, default_value)

    return result_values


def get_task_configs(cluster, is_distributed, configs, default_config, task_type):
    return get_task_values(cluster=cluster,
                           is_distributed=is_distributed,
                           values=configs,
                           default_value=default_config,
                           task_type=task_type)


def get_task_job_resources(cluster, is_distributed, resources, default_resources, task_type):
    return get_task_values(cluster=cluster,
                           is_distributed=is_distributed,
                           values=resources,
                           default_value=default_resources,
                           task_type=task_type)


def get_task_job_node_selectors(cluster,
                                is_distributed,
                                node_selectors,
                                default_node_selector,
                                task_type):
    return get_task_values(cluster=cluster,
                           is_distributed=is_distributed,
                           values=node_selectors,
                           default_value=default_node_selector,
                           task_type=task_type)


def get_task_job_tolerations(cluster,
                             is_distributed,
                             tolerations,
                             default_tolerations,
                             task_type):
    return get_task_values(cluster=cluster,
                           is_distributed=is_distributed,
                           values=tolerations,
                           default_value=default_tolerations,
                           task_type=task_type)


def get_task_job_affinities(cluster,
                            is_distributed,
                            affinities,
                            default_affinity,
                            task_type):
    return get_task_values(cluster=cluster,
                           is_distributed=is_distributed,
                           values=affinities,
                           default_value=default_affinity,
                           task_type=task_type)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function


class ExperimentBackend(object):
    NATIVE = 'native'
    KUBEFLOW = 'kubeflow'
    MPI = 'mpi'
    OTHER = 'other'

    VALUES = [NATIVE, KUBEFLOW, MPI, OTHER]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.ops.run.cmd import RunConfig, RunSchema, get_container_command_args  # noqa
from polyaxon_schemas.ops.run.op import BaseRunConfig, BaseRunSchema  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseOneOfSchema, BaseSchema


class IntervalScheduleSchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('interval'))
    start_at = fields.LocalDateTime(required=True)
    end_at = fields.LocalDateTime(allow_none=True)
    frequency = fields.Int(precision="microseconds", required=True)
    depends_on_past = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return IntervalScheduleConfig


class IntervalScheduleConfig(BaseConfig):
    SCHEMA = IntervalScheduleSchema
    IDENTIFIER = 'interval'

    def __init__(self, frequency, start_at, end_at=None, depends_on_past=None, kind=None):
        self.frequency = frequency
        self.start_at = start_at
        self.end_at = end_at
        self.depends_on_past = depends_on_past
        self.kind = kind


class CronScheduleSchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('cron'))
    start_at = fields.LocalDateTime(allow_none=True)
    end_at = fields.LocalDateTime(allow_none=True)
    cron = fields.String(required=True)
    depends_on_past = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return CronScheduleConfig


class CronScheduleConfig(BaseConfig):
    SCHEMA = CronScheduleSchema
    IDENTIFIER = 'cron'

    def __init__(self, cron, start_at=None, end_at=None, depends_on_past=None, kind=None):
        self.cron = cron
        self.start_at = start_at
        self.end_at = end_at
        self.depends_on_past = depends_on_past
        self.kind = kind


class ScheduleSchema(BaseOneOfSchema):
    TYPE_FIELD = 'kind'
    TYPE_FIELD_REMOVE = False
    SCHEMAS = {
        IntervalScheduleConfig.IDENTIFIER: IntervalScheduleSchema,
        CronScheduleConfig.IDENTIFIER: CronScheduleSchema,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields


class StrOrFct(fields.Str):
    def serialize(self, attr, obj, accessor=None, **kwargs):
        value = getattr(obj, attr)
        if hasattr(value, '__call__') and hasattr(value, '__name__'):
            return value.__name__

        return super(StrOrFct, self).serialize(attr, obj, accessor)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.base import BaseOneOfSchema
from polyaxon_schemas.registry import SchemaRegistry


class OpResolverSchema(BaseOneOfSchema):
    TYPE_FIELD = 'kind'
    TYPE_FIELD_remove = False
    SCHEMAS = SchemaRegistry(base_operations={}, polyflow_operations={})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from hestia.cached_property import cached_property

from polyaxon_schemas.ops.notebook import NotebookConfig
from polyaxon_schemas.specs import kinds
from polyaxon_schemas.specs.base import BaseRunSpecification, BaseSpecification


class NotebookSpecification(BaseRunSpecification):
    """The polyaxonfile specification for notebooks.

    SECTIONS:
        VERSION: defines the version of the file to be parsed and validated.
        LOGGING: defines the logging
        TAGS: defines the tags
        ENVIRONMENT: defines the run environment for experiment.
        BUILD: defines the build step where the user can set a docker image definition
    """
    _SPEC_KIND = kinds.NOTEBOOK

    REQUIRED_SECTIONS = BaseRunSpecification.REQUIRED_SECTIONS + (
        BaseSpecification.BUILD,
    )

    ENVIRONMENT_CONFIG = NotebookConfig
    CONFIG = NotebookConfig

    @cached_property
    def backend(self):
        return self.config.backend



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseOneOfSchema, BaseSchema
from polyaxon_schemas.ops.group.metrics import Optimization


class MedianStoppingPolicySchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('median'))
    evaluation_interval = fields.Int()

    @staticmethod
    def schema_config():
        return MedianStoppingPolicyConfig


class MedianStoppingPolicyConfig(BaseConfig):
    """
    Early stopping with median stopping, this policy computes running medians across
    all experiments and stops those whose best performance is worse than the median
    of the running experiments.

    Args:
        evaluation_interval: `int`. Frequency for applying the policy.
    """

    IDENTIFIER = 'median'
    SCHEMA = MedianStoppingPolicySchema

    def __init__(self,
                 evaluation_interval,
                 kind='median'):
        self.kind = kind
        self.evaluation_interval = evaluation_interval


class AverageStoppingPolicySchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('average'))
    evaluation_interval = fields.Int()

    @staticmethod
    def schema_config():
        return AverageStoppingPolicyConfig


class AverageStoppingPolicyConfig(BaseConfig):
    """
    Early stopping with average stopping, this policy computes running averages across
    all experiments and stops those whose best performance is worse than the median
    of the running experiments.

    Args:
        evaluation_interval: `int`. Frequency for applying the policy.
    """

    IDENTIFIER = 'average'
    SCHEMA = AverageStoppingPolicySchema

    def __init__(self,
                 evaluation_interval,
                 kind='average'):
        self.kind = kind
        self.evaluation_interval = evaluation_interval


class TruncationStoppingPolicySchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('truncation'))
    percent = fields.Int()
    evaluation_interval = fields.Int()

    @staticmethod
    def schema_config():
        return TruncationStoppingPolicyConfig


class TruncationStoppingPolicyConfig(BaseConfig):
    """
    Early stopping with truncation stopping, this policy stops a percentage of
    all running experiments at every evaluation.

    Args:
        percent: `int`. e.g. 1 - 99. The percentage of experiments to stop,
                 at each evaluation interval.
        evaluation_interval: `int`. Frequency for applying the policy.
    """

    IDENTIFIER = 'truncation'
    SCHEMA = TruncationStoppingPolicySchema

    def __init__(self,
                 percent,
                 evaluation_interval,
                 kind='truncation'):
        self.kind = kind
        self.percent = percent
        self.evaluation_interval = evaluation_interval


class StoppingPolicySchema(BaseOneOfSchema):
    TYPE_FIELD = 'kind'
    TYPE_FIELD_REMOVE = False
    SCHEMAS = {
        MedianStoppingPolicyConfig.IDENTIFIER: MedianStoppingPolicySchema,
        AverageStoppingPolicyConfig.IDENTIFIER: AverageStoppingPolicySchema,
        TruncationStoppingPolicyConfig.IDENTIFIER: TruncationStoppingPolicySchema,
    }


class EarlyStoppingSchema(BaseSchema):
    metric = fields.Str()
    value = fields.Float()
    optimization = fields.Str(allow_none=True, validate=validate.OneOf(Optimization.VALUES))
    policy = fields.Nested(StoppingPolicySchema, allow_none=True)

    @staticmethod
    def schema_config():
        return EarlyStoppingConfig


class EarlyStoppingConfig(BaseConfig):
    """
    Early stopping metric config.

    Args:
        metric: `str`. The metric to use for early stopping.
        value: `float`. The metric value to use for the condition.
        optimization: `string`. The optimization to do: maximize or minimize.
        policy: `Dict`. Optional, the termination policy to use.
    """
    SCHEMA = EarlyStoppingSchema
    IDENTIFIER = 'early_stopping_metric'

    def __init__(self,
                 metric,
                 value=None,
                 optimization=Optimization.MAXIMIZE,
                 policy=None):
        self.metric = metric
        self.value = value
        self.optimization = optimization
        self.policy = policy



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from marshmallow import fields


class IndexedDict(fields.Dict):
    def _validated(self, value):
        """Check the dict has an index or raise a :exc:`ValidationError` if an error occurs."""
        if not (isinstance(value, Mapping) or 'index' in value):
            self.fail('invalid')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError

from polyaxon_schemas.polyflow.conditions import (
    ConditionSchema,
    OutputsConditionConfig,
    StatusConditionConfig
)


class TestCondsConfigs(TestCase):
    def test_status_cond(self):
        config_dict = {'foo': 'bar', 'op': 'foo', 'trigger': 'done'}
        with self.assertRaises(ValidationError):
            StatusConditionConfig.from_dict(config_dict)

        config_dict = {'kind': 'foo', 'op': 'foo', 'trigger': 'done'}
        with self.assertRaises(ValidationError):
            StatusConditionConfig.from_dict(config_dict)

        config_dict = {
            'op': 'foo',
            'trigger': 'done'
        }
        StatusConditionConfig.from_dict(config_dict)

        config_dict = {
            'kind': 'status',
            'op': 'foo',
            'trigger': 'done'
        }
        StatusConditionConfig.from_dict(config_dict)

    def test_outputs_cond(self):
        config_dict = {'op': 'foo', 'exp': 'done', 'params': ['op1.done', 'foo']}
        with self.assertRaises(ValidationError):
            OutputsConditionConfig.from_dict(config_dict)

        config_dict = {'kind': 'foo',
                       'op': 'foo',
                       'exp': 'eq',
                       'params': [['op1.done', 'foo']]}
        with self.assertRaises(ValidationError):
            OutputsConditionConfig.from_dict(config_dict)

        config_dict = {'op': 'foo',
                       'exp': 'eq',
                       'params': ['op1.done', 'foo']}
        with self.assertRaises(ValidationError):
            OutputsConditionConfig.from_dict(config_dict)

        config_dict = {
            'op': 'foo',
            'exp': 'eq',
            'params': [['op1.done', 'foo']]
        }
        OutputsConditionConfig.from_dict(config_dict)

        config_dict = {
            'kind': 'outputs',
            'op': 'foo',
            'exp': 'eq',
            'params': [['op1.done', 'foo']]
        }
        OutputsConditionConfig.from_dict(config_dict)

    def test_conds(self):
        configs = [
            {
                'kind': 'status',
                'op': 'foo',
                'trigger': 'done'
            },
            {
                'kind': 'outputs',
                'op': 'foo',
                'exp': 'eq',
                'params': [['op1.done', 'foo']]
            }
        ]

        ConditionSchema().load(configs, many=True)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from tests.utils import assert_equal_dict

from polyaxon_schemas.ml.hooks import StepLoggingTensorHookConfig
from polyaxon_schemas.ml.processing.pipelines import TFRecordSequencePipelineConfig
from polyaxon_schemas.ml.train import TrainConfig


class TestTrainConfigs(TestCase):
    def test_train_config(self):
        config_dict = {
            'data_pipeline': TFRecordSequencePipelineConfig(
                data_files=['~/data_file'],
                meta_data_file='~/meta_data_file',
                shuffle=True,
                num_epochs=10,
                batch_size=64).to_schema(),
            'steps': 300,
            'hooks': [
                StepLoggingTensorHookConfig(['Dense_1', 'Conv2D_4'], every_n_iter=100).to_schema()
            ]
        }
        config = TrainConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import NAME_REGEX, BaseConfig, BaseSchema
from polyaxon_schemas.ops.build_job import BuildSchema
from polyaxon_schemas.ops.experiment.environment import ExperimentEnvironmentSchema
from polyaxon_schemas.ops.logging import LoggingSchema
from polyaxon_schemas.polyflow.conditions import ConditionSchema
from polyaxon_schemas.polyflow.template_ref import TemplateRefSchema
from polyaxon_schemas.polyflow.trigger_policies import TriggerPolicy


class OpSchema(BaseSchema):
    version = fields.Int(allow_none=True)
    kind = fields.Str(allow_none=True)
    template = fields.Nested(TemplateRefSchema, allow_none=True)
    logging = fields.Nested(LoggingSchema, allow_none=True)
    name = fields.Str(validate=validate.Regexp(regex=NAME_REGEX), allow_none=True)
    description = fields.Str(allow_none=True)
    tags = fields.List(fields.Str(), allow_none=True)
    params = fields.Raw(allow_none=True)
    environment = fields.Nested(ExperimentEnvironmentSchema, allow_none=True)
    build = fields.Nested(BuildSchema, allow_none=True)
    concurrency = fields.Int(allow_none=True)
    dependencies = fields.List(fields.Str(), allow_none=True)
    trigger = fields.Str(allow_none=True, validate=validate.OneOf(TriggerPolicy.VALUES))
    conditions = fields.Nested(ConditionSchema, allow_none=True)
    max_retries = fields.Int(allow_none=True)
    retry_delay = fields.Int(allow_none=True)
    retry_exp_backoff = fields.Bool(allow_none=True)
    max_retry_delay = fields.Int(allow_none=True)
    skip_on_upstream_skip = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return OpConfig


class OpConfig(BaseConfig):
    SCHEMA = OpSchema
    IDENTIFIER = 'op'
    REDUCED_ATTRIBUTES = [
        'version',
        'kind',
        'template',
        'logging',
        'name',
        'description',
        'tags',
        'params',
        'environment',
        'build',
        'concurrency',
        'dependencies',
        'trigger',
        'conditions',
        'max_retries',
        'retry_delay',
        'retry_exp_backoff',
        'max_retry_delay',
        'skip_on_upstream_skip',
    ]

    def __init__(self,
                 version=None,
                 kind=None,
                 template=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 params=None,
                 environment=None,
                 build=None,
                 concurrency=None,
                 dependencies=None,
                 trigger=None,
                 conditions=None,
                 max_retries=None,
                 retry_delay=None,
                 retry_exp_backoff=None,
                 max_retry_delay=None,
                 skip_on_upstream_skip=None):
        self.version = version
        self.kind = kind
        self.template = template
        self.logging = logging
        self.name = name
        self.description = description
        self.tags = tags
        self.environment = environment
        self.params = params
        self.build = build
        self.concurrency = concurrency
        self.dependencies = dependencies
        self.trigger = trigger
        self.conditions = conditions
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        self.retry_exp_backoff = retry_exp_backoff
        self.max_retry_delay = max_retry_delay
        self.skip_on_upstream_skip = skip_on_upstream_skip



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from mock import MagicMock

from polyaxon_schemas.ops.run import RunConfig, get_container_command_args


class TestContainerCmdArgs(TestCase):
    def setUp(self):
        self.cmd1 = MagicMock()
        self.cmd1.cmd = "video_prediction_train --model=DNA --num_masks=1"
        self.cmd2 = MagicMock()
        self.cmd2.cmd = "/bin/bash run.sh"
        self.cmd3 = MagicMock()
        self.cmd3.cmd = """
                python3 model.py --batch_size={{ batch_size }} \
                                 --num_steps={{ num_steps }} \
                                 --learning_rate={{ learning_rate }} \
                                 --dropout={{ dropout }} \
                                 --num_epochs={{ num_epochs }} \
                                 --activation={{ activation }}
                """
        self.cmd4 = MagicMock()
        self.cmd4.cmd = """
                video_prediction_train --model=DNA --num_masks=1 && \
                python3 model.py --batch_size={{ batch_size }} \
                                 --num_steps={{ num_steps }} \
                                 --learning_rate={{ learning_rate }} \
                                 --dropout={{ dropout }} \
                                 --num_epochs={{ num_epochs }} \
                                 --activation={{ activation }}
                """
        self.cmd5 = MagicMock()
        self.cmd5.cmd = """
                        video_prediction_train --model=DNA --num_masks=1 ; \
                        python3 model.py --batch_size={{ batch_size }} \
                                         --num_steps={{ num_steps }} \
                                         --learning_rate={{ learning_rate }} \
                                         --dropout={{ dropout }} \
                                         --num_epochs={{ num_epochs }} \
                                         --activation={{ activation }}
                        """
        self.cmd6 = MagicMock()
        self.cmd6.cmd = """
                                video_prediction_train --model=DNA --num_masks=1 || \
                                python3 model.py --batch_size={{ batch_size }} \
                                                 --num_steps={{ num_steps }} \
                                                 --learning_rate={{ learning_rate }} \
                                                 --dropout={{ dropout }} \
                                                 --num_epochs={{ num_epochs }} \
                                                 --activation={{ activation }}
                                """

        self.list_cmd1 = MagicMock()
        self.list_cmd1.cmd = ["video_prediction_train --model=DNA --num_masks=1",
                              "video_prediction_train --model=NEW --num_masks=10"]
        self.list_cmd2 = MagicMock()
        self.list_cmd2.cmd = ["/bin/bash run1.sh", "/bin/bash run2.sh", "/bin/bash run3.sh"]
        self.list_cmd3 = MagicMock()
        self.list_cmd3.cmd = [
            "/bin/bash run1.sh",
            """
            python3 model.py --batch_size={{ batch_size }} \
                             --num_steps={{ num_steps }} \
                             --learning_rate={{ learning_rate }} \
                             --dropout={{ dropout }} \
                             --num_epochs={{ num_epochs }} \
                             --activation={{ activation }}
            """]
        self.list_cmd4 = MagicMock()
        self.list_cmd4.cmd = [
            "/bin/bash run1.sh",
            """
            video_prediction_train --model=DNA --num_masks=1 && \
            python3 model.py --batch_size={{ batch_size }} \
                             --num_steps={{ num_steps }} \
                             --learning_rate={{ learning_rate }} \
                             --dropout={{ dropout }} \
                             --num_epochs={{ num_epochs }} \
                             --activation={{ activation }}
            """]
        self.list_cmd5 = MagicMock()
        self.list_cmd5.cmd = [
            "/bin/bash run1.sh",
            """
            video_prediction_train --model=DNA --num_masks=1 ; \
            python3 model.py --batch_size={{ batch_size }} \
                             --num_steps={{ num_steps }} \
                             --learning_rate={{ learning_rate }} \
                             --dropout={{ dropout }} \
                             --num_epochs={{ num_epochs }} \
                             --activation={{ activation }}
            """]
        self.list_cmd6 = MagicMock()
        self.list_cmd6.cmd = [
            "/bin/bash run1.sh",
            """
            video_prediction_train --model=DNA --num_masks=1 || \
            python3 model.py --batch_size={{ batch_size }} \
                             --num_steps={{ num_steps }} \
                             --learning_rate={{ learning_rate }} \
                             --dropout={{ dropout }} \
                             --num_epochs={{ num_epochs }} \
                             --activation={{ activation }}
            """]

    def test_get_container_command_args_cmd(self):
        assert get_container_command_args(self.cmd1) == (
            ["/bin/bash", "-c"],
            ["video_prediction_train --model=DNA --num_masks=1"])
        assert get_container_command_args(self.cmd2) == (["/bin/bash", "-c"], ["/bin/bash run.sh"])
        assert get_container_command_args(self.cmd3) == (
            ["/bin/bash", "-c"],
            ["python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

        assert get_container_command_args(self.cmd4) == (
            ["/bin/bash", "-c"],
            ["video_prediction_train --model=DNA --num_masks=1 && "
             "python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

        assert get_container_command_args(self.cmd5) == (
            ["/bin/bash", "-c"],
            ["video_prediction_train --model=DNA --num_masks=1 ; "
             "python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

        assert get_container_command_args(self.cmd6) == (
            ["/bin/bash", "-c"],
            ["video_prediction_train --model=DNA --num_masks=1 || "
             "python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

    def test_get_container_command_args_list_cmd(self):
        assert get_container_command_args(self.list_cmd1) == (
            ["/bin/bash", "-c"],
            ["video_prediction_train --model=DNA --num_masks=1 && "
             "video_prediction_train --model=NEW --num_masks=10"])
        assert get_container_command_args(self.list_cmd2) == (
            ["/bin/bash", "-c"],
            ["/bin/bash run1.sh && /bin/bash run2.sh && /bin/bash run3.sh"])
        assert get_container_command_args(self.list_cmd3) == (
            ["/bin/bash", "-c"],
            ["/bin/bash run1.sh && python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

        assert get_container_command_args(self.list_cmd4) == (
            ["/bin/bash", "-c"],
            ["/bin/bash run1.sh && video_prediction_train --model=DNA --num_masks=1 && "
             "python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

        assert get_container_command_args(self.list_cmd5) == (
            ["/bin/bash", "-c"],
            ["/bin/bash run1.sh && video_prediction_train --model=DNA --num_masks=1 ; "
             "python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])

        assert get_container_command_args(self.list_cmd6) == (
            ["/bin/bash", "-c"],
            ["/bin/bash run1.sh && video_prediction_train --model=DNA --num_masks=1 || "
             "python3 model.py "
             "--batch_size={{ batch_size }} "
             "--num_steps={{ num_steps }} "
             "--learning_rate={{ learning_rate }} "
             "--dropout={{ dropout }} "
             "--num_epochs={{ num_epochs }} "
             "--activation={{ activation }}"])


class TestRunConfigs(TestCase):
    def test_exec_config_with_str_cmd(self):
        config_dict = {
            'cmd': 'python t2t-trainer '
                   '--generate_data '
                   '--data_dir=~/t2t_data '
                   '--problems=translate_ende_wmt32k '
                   '--model=transformer '
                   '--hparams_set=transformer_base_single_gpu '
                   '--output_dir=~/t2t_train/base',
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base']
        )

        config_dict = {
            'cmd': 'foo && python t2t-trainer '
                   '--generate_data '
                   '--data_dir=~/t2t_data '
                   '--problems=translate_ende_wmt32k '
                   '--model=transformer '
                   '--hparams_set=transformer_base_single_gpu '
                   '--output_dir=~/t2t_train/base',
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['foo && python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base']
        )

        config_dict = {
            'cmd': 'foo; python t2t-trainer '
                   '--generate_data '
                   '--data_dir=~/t2t_data '
                   '--problems=translate_ende_wmt32k '
                   '--model=transformer '
                   '--hparams_set=transformer_base_single_gpu '
                   '--output_dir=~/t2t_train/base',
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['foo; python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base',]
        )

        config_dict = {
            'cmd': 'foo || python t2t-trainer '
                   '--generate_data '
                   '--data_dir=~/t2t_data '
                   '--problems=translate_ende_wmt32k '
                   '--model=transformer '
                   '--hparams_set=transformer_base_single_gpu '
                   '--output_dir=~/t2t_train/base',
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_exec_config_with_list_cmd(self):
        config_dict = {
            'cmd': [
                'foo',
                'python t2t-trainer '
                '--generate_data '
                '--data_dir=~/t2t_data '
                '--problems=translate_ende_wmt32k '
                '--model=transformer '
                '--hparams_set=transformer_base_single_gpu '
                '--output_dir=~/t2t_train/base',
            ]
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['foo && python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base']
        )

        config_dict = {
            'cmd': [
                'foo && another_cmd && python t2t-trainer '
                '--generate_data '
                '--data_dir=~/t2t_data '
                '--problems=translate_ende_wmt32k '
                '--model=transformer '
                '--hparams_set=transformer_base_single_gpu '
                '--output_dir=~/t2t_train/base',
            ]
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['foo && another_cmd && python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base']
        )

        config_dict = {
            'cmd': [
                'foo && another_cmd; python t2t-trainer '
                '--generate_data '
                '--data_dir=~/t2t_data '
                '--problems=translate_ende_wmt32k '
                '--model=transformer '
                '--hparams_set=transformer_base_single_gpu '
                '--output_dir=~/t2t_train/base',
            ]
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['foo && another_cmd; python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base',]
        )

        config_dict = {
            'cmd': [
                'foo && another_cmd; || python t2t-trainer '
                '--generate_data '
                '--data_dir=~/t2t_data '
                '--problems=translate_ende_wmt32k '
                '--model=transformer '
                '--hparams_set=transformer_base_single_gpu '
                '--output_dir=~/t2t_train/base',
            ]
        }
        config = RunConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.get_container_cmd() == (
            ["/bin/bash", "-c"],
            ['foo && another_cmd; || python t2t-trainer '
             '--generate_data '
             '--data_dir=~/t2t_data '
             '--problems=translate_ende_wmt32k '
             '--model=transformer '
             '--hparams_set=transformer_base_single_gpu '
             '--output_dir=~/t2t_train/base']
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from polyaxon_schemas.ml.bridges import LatentBridgeConfig, NoOpBridgeConfig


class TestConstraintConfigs(TestCase):
    def test_latent_bridge_config(self):
        config_dict = {
            'state_size': [2, 2],
            'latent_dim': 1,
            'mean': 1.,
            'stddev': 0.,
            'name': 'latent',
        }
        config = LatentBridgeConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_noop_bridge_config(self):
        config_dict = {
            'state_size': [2, 2],
            'name': 'noop',
        }
        config = NoOpBridgeConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseOneOfSchema, BaseSchema
from polyaxon_schemas.polyflow.trigger_policies import ExpressionTriggerPolicy, StatusTriggerPolicy


class StatusConditionSchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('status'))
    op = fields.Str()
    trigger = fields.Str(allow_none=True, validate=validate.OneOf(StatusTriggerPolicy.VALUES))

    @staticmethod
    def schema_config():
        return StatusConditionConfig


class StatusConditionConfig(BaseConfig):
    SCHEMA = StatusConditionSchema
    IDENTIFIER = 'status'

    def __init__(self, op, trigger, kind=None):
        self.op = op
        self.trigger = trigger
        self.kind = kind


class OutputsConditionSchema(BaseSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('outputs'))
    op = fields.Str()
    exp = fields.Str(allow_none=True, validate=validate.OneOf(ExpressionTriggerPolicy.VALUES))
    params = env_vars = fields.List(fields.List(fields.Raw(), validate=validate.Length(equal=2)))

    @staticmethod
    def schema_config():
        return OutputsConditionConfig


class OutputsConditionConfig(BaseConfig):
    SCHEMA = OutputsConditionSchema
    IDENTIFIER = 'outputs'

    def __init__(self, op, exp, params, kind=None):
        self.op = op
        self.exp = exp
        self.params = params
        self.kind = kind


class ConditionSchema(BaseOneOfSchema):
    TYPE_FIELD = 'kind'
    TYPE_FIELD_REMOVE = False
    SCHEMAS = {
        StatusConditionConfig.IDENTIFIER: StatusConditionSchema,
        OutputsConditionConfig.IDENTIFIER: OutputsConditionSchema,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import re
import warnings

from marshmallow import ValidationError, fields, validate, validates_schema

from polyaxon_schemas.base import NAME_REGEX, BaseConfig, BaseSchema
from polyaxon_schemas.ops import params as ops_params
from polyaxon_schemas.ops.environments.pods import EnvironmentSchema
from polyaxon_schemas.ops.io import IOSchema
from polyaxon_schemas.ops.logging import LoggingSchema

PARAM_REGEX = re.compile(r'{{\s*([^\s]*)\s*}}')


def validate_declarations(values):
    if values.get('declarations') and values.get('params'):
        raise ValidationError('You should only use `params`.')

    if values.get('declarations'):
        warnings.warn(
            'The `declarations` parameter is deprecated and will be removed in next release, '
            'please use `params` instead',
            DeprecationWarning)
        values['params'] = values.pop('declarations')

    return values


class BaseOpSchema(BaseSchema):
    version = fields.Int(allow_none=True)
    kind = fields.Str(allow_none=True)
    logging = fields.Nested(LoggingSchema, allow_none=True)
    name = fields.Str(validate=validate.Regexp(regex=NAME_REGEX), allow_none=True)
    description = fields.Str(allow_none=True)
    tags = fields.List(fields.Str(), allow_none=True)
    environment = fields.Nested(EnvironmentSchema, allow_none=True)
    params = fields.Raw(allow_none=True)
    declarations = fields.Raw(allow_none=True)
    inputs = fields.Nested(IOSchema, allow_none=True, many=True)
    outputs = fields.Nested(IOSchema, allow_none=True, many=True)

    @staticmethod
    def schema_config():
        return BaseOpConfig

    @validates_schema
    def validate_declarations(self, values):
        validate_declarations(values)

    @validates_schema
    def validate_params(self, values):
        ops_params.validate_params(params=values.get('params'),
                                   inputs=values.get('inputs'),
                                   outputs=values.get('outputs'),
                                   is_template=True,
                                   is_run=True)


class BaseOpConfig(BaseConfig):
    SCHEMA = BaseOpSchema
    IDENTIFIER = 'operation'
    REDUCED_ATTRIBUTES = [
        'version',
        'kind',
        'logging',
        'name',
        'description',
        'tags',
        'environment',
        'params',
        'inputs',
        'outputs',
    ]

    def __init__(self,
                 version=None,
                 kind=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 environment=None,
                 params=None,
                 declarations=None,
                 inputs=None,
                 outputs=None):
        self.version = version
        self.kind = kind
        self.logging = logging
        self.name = name
        self.description = description
        self.tags = tags
        self.environment = environment
        validate_declarations({'params': params, 'declarations': declarations})
        self.params = params or declarations
        self._validated_params = ops_params.validate_params(params=self.params,
                                                            inputs=inputs,
                                                            outputs=outputs,
                                                            is_template=True,
                                                            is_run=True)
        self.inputs = inputs
        self.outputs = outputs

    def get_params(self, context):
        """Return all params: Merge params if passed, if not default values."""
        if not self._validated_params:
            return self.params

        params = {}
        for param in self._validated_params:
            if not param.entity_ref:
                params[param.name] = param.value
            else:
                params[param.name] = context[param.value.replace('.', '__')]

        return params

    def get_params_with_refs(self):
        return [param for param in self._validated_params if param.entity_ref]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from collections import namedtuple

from marshmallow import fields, validate

from polyaxon_schemas.base import NAME_REGEX
from polyaxon_schemas.exceptions import PolyaxonSchemaError
from polyaxon_schemas.ops import params as ops_params
from polyaxon_schemas.ops.logging import LoggingSchema
from polyaxon_schemas.polyflow.executable import ExecutableConfig, ExecutableSchema
from polyaxon_schemas.polyflow.ops import OpSchema
from polyaxon_schemas.polyflow.schedule import ScheduleSchema
from polyaxon_schemas.polyflow.template import TemplateSchema


class DagOpSpec(namedtuple("DagOpSpec", "op upstream downstream")):

    def items(self):
        return self._asdict().items()

    def set_op(self, op):
        return self._replace(op=op)


class PipelineSchema(ExecutableSchema):
    version = fields.Int(allow_none=True)
    kind = fields.Str(allow_none=True, validate=validate.Equal('pipeline'))
    name = fields.Str(validate=validate.Regexp(regex=NAME_REGEX), allow_none=True)
    logging = fields.Nested(LoggingSchema, allow_none=True)
    tags = fields.List(fields.Str(), allow_none=True)
    backend = fields.Str(allow_none=True)
    ops = fields.Nested(OpSchema, many=True)
    templates = fields.Nested(TemplateSchema, many=True)
    schedule = fields.Nested(ScheduleSchema, allow_none=True)
    concurrency = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return PipelineConfig


class PipelineConfig(ExecutableConfig):
    SCHEMA = PipelineSchema
    IDENTIFIER = 'pipeline'
    REDUCED_ATTRIBUTES = ExecutableConfig.REDUCED_ATTRIBUTES + [
        'kind',
        'version',
        'name',
        'logging',
        'tags',
        'backend',
        'ops',
        'concurrency',
        'schedule',
        'templates',
    ]

    def __init__(self,
                 kind=None,
                 version=None,
                 name=None,
                 logging=None,
                 tags=None,
                 backend=None,
                 ops=None,
                 concurrency=None,
                 schedule=None,
                 templates=None,
                 execute_at=None,
                 timeout=None):
        super(PipelineConfig, self).__init__(execute_at=execute_at, timeout=timeout)
        self.kind = kind
        self.version = version
        self.name = name
        self.logging = logging
        self.tags = tags
        self.backend = backend
        self.ops = ops
        self.concurrency = concurrency
        self.schedule = schedule
        self.templates = templates
        self._dag = {}  # OpName -> DagOpSpec
        self._template_by_names = {}  # TemplateName -> Template
        self._op_template_mapping = {}  # OpName -> TemplateName
        self._context = {}  # Ops output names -> IOTypes

    @property
    def dag(self):
        return self._dag

    def validate_dag(self, dag=None):
        dag = dag or self.dag
        orphan_ops = self.get_orphan_ops(dag=dag)
        if orphan_ops:
            raise PolyaxonSchemaError('Pipeline has a non valid dag, the dag contains orphan ops: '
                                      '`{}`'.format(orphan_ops))
        self.sort_topologically(dag=dag)

    def _get_op_upstream(self, op):
        upstream = set(op.dependencies) if op.dependencies else set([])

        if not op.params:
            return upstream

        for param in op.params:
            param_ref = ops_params.get_param(name=param, value=op.params[param], iotype=None)
            if param_ref and param_ref.entity == ops_params.OPS:
                upstream.add(param_ref.entity_ref)

        return upstream

    def _set_dag_op(self, op_name, op=None, upstream=None, downstream=None):
        upstream = set(upstream) if upstream else set([])
        downstream = set(downstream) if downstream else set([])
        if op_name in self._dag:
            if op and self._dag[op_name].op is None:
                self._dag[op_name] = self._dag[op_name].set_op(op)
            self._dag[op_name].upstream.update(upstream)
            self._dag[op_name].downstream.update(downstream)
        else:
            self._dag[op_name] = DagOpSpec(op=op, upstream=upstream, downstream=downstream)

    def _process_op(self, op):
        upstream = self._get_op_upstream(op=op)
        self._set_dag_op(op_name=op.name, op=op, upstream=upstream, downstream=None)
        for op_name in upstream:
            self._set_dag_op(op_name=op_name, downstream=[op.name])

    def process_dag(self):
        for op in self.ops or []:
            self._process_op(op)

    def add_op(self, op):
        self.ops = self.ops or []
        self.ops.append(op)
        self._process_op(op)

    def add_ops(self, ops):
        for op in ops:
            self.add_op(op)

    def get_independent_ops(self, dag=None):
        """Get a list of all node in the graph with no dependencies."""
        dag = dag or self.dag
        ops = set(six.iterkeys(dag))
        dependent_nodes = {op_downstream for op in six.itervalues(dag)
                           for op_downstream in op.downstream}
        return ops - dependent_nodes

    def get_orphan_ops(self, dag=None):
        """Get orphan ops for given dag."""
        dag = dag or self.dag
        independent_ops = self.get_independent_ops(dag)
        return {op for op in independent_ops if dag[op].op is None}

    def sort_topologically(self, dag=None, flatten=False):
        """Sort the dag breath first topologically.

        Only the nodes inside the dag are returned, i.e. the nodes that are also keys.

        Returns:
             a topological ordering of the DAG.
        Raises:
             an error if this is not possible (graph is not valid).
        """
        def get_independent_ops():
            if current_independent_ops:
                return current_independent_ops.pop()
            current_independent_ops.update(next_independent_ops)
            next_independent_ops.clear()
            sorted_ops.append(current_sorted_ops[:])
            del current_sorted_ops[:]  # in python3 it should use .clear()
            if current_independent_ops:
                return current_independent_ops.pop()

        dag = dag or self.dag
        visited_ops = set()
        next_independent_ops = set()
        current_sorted_ops = []
        sorted_ops = []
        current_independent_ops = self.get_independent_ops(dag)
        op = get_independent_ops()
        while op:
            current_sorted_ops.append(op)
            visited_ops.add(op)
            downstream_ops = dag[op].downstream.copy()
            while downstream_ops:
                downstream_op = downstream_ops.pop()

                if downstream_op not in dag:
                    continue

                if downstream_op in visited_ops:
                    continue

                if not dag[downstream_op].upstream - visited_ops:
                    next_independent_ops.add(downstream_op)

            op = get_independent_ops()

        flatten_sorted_ops = [i for il in sorted_ops for i in il]
        if len(flatten_sorted_ops) != len(dag.keys()):
            raise PolyaxonSchemaError("Pipeline's graph is not acyclic.")
        return flatten_sorted_ops if flatten else sorted_ops

    def process_templates(self):
        if not self.templates:
            raise PolyaxonSchemaError('Pipeline is not valid, '
                                      'it has no templates to validate operations.')

        if not self.ops:
            raise PolyaxonSchemaError('Pipeline is not valid, '
                                      'it has no ops to validate operations.')

        for template in self.templates:
            template_name = template.name
            if template_name in self._template_by_names:
                raise PolyaxonSchemaError(
                    'Pipeline has multiple templates with the same name `{}`'.format(template_name))
            self._template_by_names[template_name] = template

        for op in self.ops:
            op_name = op.name
            template_name = op.template.name
            if op_name in self._op_template_mapping:
                raise PolyaxonSchemaError(
                    'Pipeline has multiple ops with the same name `{}`'.format(op_name))

            if template_name not in self._template_by_names:
                raise PolyaxonSchemaError(
                    'Pipeline op with name `{}` requires a template with name `{}`, '
                    'which is not defined on this pipeline.'.format(op_name, template_name))

            self._op_template_mapping[op_name] = template_name
            outputs = self._template_by_names[template_name].outputs
            if not outputs:
                continue
            for output in outputs:
                self._context['ops.{}.outputs.{}'.format(op_name, output.name)] = output

        for op in self.ops:
            template = self._template_by_names[self._op_template_mapping[op.name]]
            ops_params.validate_params(params=op.params,
                                       inputs=template.inputs,
                                       outputs=template.outputs,
                                       context=self._context,
                                       is_template=False,
                                       is_run=False)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import EXCLUDE, fields

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema
from polyaxon_schemas.fields import Tensor


class BaseLossSchema(BaseSchema):
    input_layer = Tensor(allow_none=True)
    output_layer = Tensor(allow_none=True)
    weights = fields.Float(default=1.0, missing=1.0)
    name = fields.Str(allow_none=True)
    collect = fields.Bool(default=True, missing=True)

    class Meta:
        unknown = EXCLUDE
        ordered = True


class BaseLossConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['input_layer', 'output_layer', 'name']

    def __init__(self, input_layer=None, output_layer=None, weights=1.0, name=None, collect=True):
        self.input_layer = input_layer
        self.output_layer = output_layer
        self.weights = weights
        self.name = name
        self.collect = collect


class AbsoluteDifferenceSchema(BaseLossSchema):

    @staticmethod
    def schema_config():
        return AbsoluteDifferenceConfig


class AbsoluteDifferenceConfig(BaseLossConfig):
    """Adds an Absolute Difference loss to the training procedure.

    `weights` acts as a coefficient for the loss. If a scalar is provided, then
    the loss is simply scaled by the given value. If `weights` is a `Tensor` of
    shape `[batch_size]`, then the total loss for each sample of the batch is
    rescaled by the corresponding element in the `weights` vector. If the shape of
    `weights` matches the shape of `predictions`, then the loss of each
    measurable element of `predictions` is scaled by the corresponding value of
    `weights`.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: AbsoluteDifference
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        AbsoluteDifference:
          input_layer: labels
          output_layer: dense_out
      # other model properties
    ```
    """
    IDENTIFIER = 'AbsoluteDifference'
    SCHEMA = AbsoluteDifferenceSchema


class MeanSquaredErrorSchema(BaseLossSchema):

    @staticmethod
    def schema_config():
        return MeanSquaredErrorConfig


class MeanSquaredErrorConfig(BaseLossConfig):
    """Computes Mean Square Loss.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: MeanSquaredError
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        MeanSquaredError:
          input_layer: labels
          output_layer: dense_out
      # other model properties
    ```
    """
    IDENTIFIER = 'MeanSquaredError'
    SCHEMA = MeanSquaredErrorSchema


class LogLossSchema(BaseLossSchema):
    epsilon = fields.Float(default=1e-7, missing=1e-7)

    @staticmethod
    def schema_config():
        return LogLossConfig


class LogLossConfig(BaseLossConfig):
    """Computes Huber Loss for DQN.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        epsilon: A small value.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: LogLoss
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        LogLoss:
          input_layer: labels
          output_layer: dense_out
      # other model properties
    ```
    """
    IDENTIFIER = 'LogLoss'
    SCHEMA = LogLossSchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 epsilon=1e-7,
                 name=None,
                 collect=True):
        super(LogLossConfig, self).__init__(input_layer, output_layer, weights, name, collect)
        self.epsilon = epsilon


class HuberLossSchema(BaseLossSchema):
    clip = fields.Float(default=0., missing=0.)

    @staticmethod
    def schema_config():
        return HuberLossConfig


class HuberLossConfig(BaseLossConfig):
    """Computes Huber Loss for DQN.

    [Wikipedia link](https://en.wikipedia.org/wiki/Huber_loss)
    [DeepMind link](https://sites.google.com/a/deepmind.com/dqn/)

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: HuberLoss
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        HuberLoss:
          clip: 0.2
      # other model properties
    ```
    """
    IDENTIFIER = 'HuberLoss'
    SCHEMA = HuberLossSchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 clip=0.,
                 name=None,
                 collect=True):
        super(HuberLossConfig, self).__init__(input_layer, output_layer, weights, name, collect)
        self.clip = clip


class ClippedDeltaLossSchema(BaseLossSchema):
    clip_value_min = fields.Float(default=-1., missing=-1.)
    clip_value_max = fields.Float(default=-1., missing=-1.)

    @staticmethod
    def schema_config():
        return ClippedDeltaLossConfig


class ClippedDeltaLossConfig(BaseLossConfig):
    """Computes clipped delta Loss for DQN.

    [Wikipedia link](https://en.wikipedia.org/wiki/Huber_loss)
    [DeepMind link](https://sites.google.com/a/deepmind.com/dqn/)

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        clip_value_min: default to -1.
        clip_value_max: default to 1.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: ClippedDeltaLoss
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        ClippedDeltaLoss:
          clip_value_min: -0.8
      # other model properties
    ```
    """
    IDENTIFIER = 'ClippedDeltaLoss'
    SCHEMA = ClippedDeltaLossSchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 clip_value_min=-1.,
                 clip_value_max=1.,
                 name=None,
                 collect=True):
        super(ClippedDeltaLossConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)
        self.clip_value_min = clip_value_min
        self.clip_value_max = clip_value_max


class SoftmaxCrossEntropySchema(BaseLossSchema):
    label_smoothing = fields.Float(default=0., missing=0.)

    @staticmethod
    def schema_config():
        return SoftmaxCrossEntropyConfig


class SoftmaxCrossEntropyConfig(BaseLossConfig):
    """Computes Softmax Cross entropy (softmax categorical cross entropy).

    Computes softmax cross entropy between y_pred (logits) and
    y_true (labels).

    Measures the probability error in discrete classification tasks in which
    the classes are mutually exclusive (each entry is in exactly one class).
    For example, each CIFAR-10 image is labeled with one and only one label:
    an image can be a dog or a truck, but not both.

    **WARNING:** This op expects unscaled logits, since it performs a `softmax`
    on `y_pred` internally for efficiency.  Do not call this op with the
    output of `softmax`, as it will produce incorrect results.

    `y_pred` and `y_true` must have the same shape `[batch_size, num_classes]`
    and the same dtype (either `float32` or `float64`). It is also required
    that `y_true` (labels) are binary arrays (For example, class 2 out of a
    total of 5 different classes, will be define as [0., 1., 0., 0., 0.])

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        label_smoothing: If greater than `0` then smooth the labels.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: SoftmaxCrossEntropy
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        SoftmaxCrossEntropy:
          label_smoothing: 0.1
      # other model properties
    ```
    """
    IDENTIFIER = 'SoftmaxCrossEntropy'
    SCHEMA = SoftmaxCrossEntropySchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 label_smoothing=0.,
                 name=None,
                 collect=True):
        super(SoftmaxCrossEntropyConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)
        self.label_smoothing = label_smoothing


class SigmoidCrossEntropySchema(BaseLossSchema):
    label_smoothing = fields.Float(default=0., missing=0.)

    @staticmethod
    def schema_config():
        return SigmoidCrossEntropyConfig


class SigmoidCrossEntropyConfig(BaseLossConfig):
    """Computes Sigmoid cross entropy.(binary cross entropy):

    Computes sigmoid cross entropy between y_pred (logits) and y_true
    (labels).

    Measures the probability error in discrete classification tasks in which
    each class is independent and not mutually exclusive. For instance,
    one could perform multilabel classification where a picture can contain
    both an elephant and a dog at the same time.

    For brevity, let `x = logits`, `z = targets`.  The logistic loss is

      x - x * z + log(1 + exp(-x))

    To ensure stability and avoid overflow, the implementation uses

      max(x, 0) - x * z + log(1 + exp(-abs(x)))

    `y_pred` and `y_true` must have the same type and shape.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.
        label_smoothing: If greater than `0` then smooth the labels.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: SigmoidCrossEntropy
      # other model properties
    ```

    or

    ```yaml
    model:
      # other model properties
      loss:
        SigmoidCrossEntropy:
          label_smoothing: 0.1
      # other model properties
    ```
    """
    IDENTIFIER = 'SigmoidCrossEntropy'
    SCHEMA = SigmoidCrossEntropySchema

    def __init__(self,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 label_smoothing=0.,
                 name=None,
                 collect=True):
        super(SigmoidCrossEntropyConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)
        self.label_smoothing = label_smoothing


class HingeLossSchema(BaseLossSchema):
    @staticmethod
    def schema_config():
        return HingeLossConfig


class HingeLossConfig(BaseLossConfig):
    """Hinge Loss.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: HingeLoss
      # other model properties
    ```
    """
    IDENTIFIER = 'HingeLoss'
    SCHEMA = HingeLossSchema


class CosineDistanceSchema(BaseLossSchema):
    dim = fields.Int()

    @staticmethod
    def schema_config():
        return CosineDistanceConfig


class CosineDistanceConfig(BaseLossConfig):
    """Adds a cosine-distance loss to the training procedure.

    Note that the function assumes that `predictions` and `labels` are already unit-normalized.

    WARNING: `weights` also supports dimensions of 1, but the broadcasting does
    not work as advertised, you'll wind up with weighted sum instead of weighted
    mean for any but the last dimension. This will be cleaned up soon, so please
    do not rely on the current behavior for anything but the shapes documented for
    `weights` below.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: CosineDistance
      # other model properties
    ```
    """
    IDENTIFIER = 'CosineDistance'
    SCHEMA = CosineDistanceSchema

    def __init__(self,
                 dim,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0,
                 name=None,
                 collect=True):
        self.dim = dim
        super(CosineDistanceConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)


class PoissonLossSchema(BaseLossSchema):
    @staticmethod
    def schema_config():
        return PoissonLossConfig


class PoissonLossConfig(BaseLossConfig):
    """Adds a poisson loss to the training procedure.

    Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: PoissonLoss
      # other model properties
    ```
    """
    IDENTIFIER = 'PoissonLoss'
    SCHEMA = PoissonLossSchema


class KullbackLeiberDivergenceSchema(BaseLossSchema):
    dim = fields.Int()

    @staticmethod
    def schema_config():
        return KullbackLeiberDivergenceConfig


class KullbackLeiberDivergenceConfig(BaseLossConfig):
    """Adds a Kullback leiber diverenge loss to the training procedure.

     Args:
        input_layer: The input true values layer, defaults to labels.
        output_layer: The output layer generated by the network,
            default to last layer of the network.
            If the network has multiple output, you should specify which layer to use.
        weights: Optional `Tensor` whose rank is either 0, or the same rank as
            `labels`, and must be broadcastable to `labels` (i.e., all dimensions must
            be either `1`, or the same as the corresponding `losses` dimension).
        name: operation name.
        collect: add to losses collection.

    Returns:
        A scalar `Tensor` representing the loss value.

    Raises:
        ValueError: If `predictions` shape doesn't match `labels` shape, or `weights` is `None`.

    Polyaxonfile usage:

    ```yaml
    model:
      # other model properties
      loss: KullbackLeiberDivergence
      # other model properties
    ```
    """
    IDENTIFIER = 'KullbackLeiberDivergence'
    SCHEMA = KullbackLeiberDivergenceSchema

    def __init__(self,
                 dim,
                 input_layer=None,
                 output_layer=None,
                 weights=1.0, name='KullbackLeiberDivergence',
                 collect=True):
        self.dim = dim
        super(KullbackLeiberDivergenceConfig, self).__init__(
            input_layer, output_layer, weights, name, collect)


class LossSchema(BaseMultiSchema):
    __multi_schema_name__ = 'loss'
    __configs__ = {
        AbsoluteDifferenceConfig.IDENTIFIER: AbsoluteDifferenceConfig,
        MeanSquaredErrorConfig.IDENTIFIER: MeanSquaredErrorConfig,
        LogLossConfig.IDENTIFIER: LogLossConfig,
        HuberLossConfig.IDENTIFIER: HuberLossConfig,
        ClippedDeltaLossConfig.IDENTIFIER: ClippedDeltaLossConfig,
        SoftmaxCrossEntropyConfig.IDENTIFIER: SoftmaxCrossEntropyConfig,
        SigmoidCrossEntropyConfig.IDENTIFIER: SigmoidCrossEntropyConfig,
        HingeLossConfig.IDENTIFIER: HingeLossConfig,
        CosineDistanceConfig.IDENTIFIER: CosineDistanceConfig,
        KullbackLeiberDivergenceConfig.IDENTIFIER: KullbackLeiberDivergenceConfig,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.ops.io.io import IOConfig, IOSchema  # noqa
from polyaxon_schemas.ops.io.types import IOTypes  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from polyaxon_schemas.fields.dict_or_str import DictOrStr  # noqa
from polyaxon_schemas.fields.float_or_str import FloatOrStr  # noqa
from polyaxon_schemas.fields.indexed_dict import IndexedDict  # noqa
from polyaxon_schemas.fields.int_or_str import IntOrStr  # noqa
from polyaxon_schemas.fields.obj_list_obj import ObjectOrListObject  # noqa
from polyaxon_schemas.fields.str_or_fct import StrOrFct  # noqa
from polyaxon_schemas.fields.tensor import Tensor  # noqa
from polyaxon_schemas.fields.uuids import UUID  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from marshmallow import ValidationError, fields, validates_schema

from polyaxon_schemas.base import BaseConfig, BaseSchema


def validate_template(**kwargs):
    if len([i for i in six.itervalues(kwargs) if i is not None and i != '']) != 1:
        raise ValidationError(
            'Template requires one and only one param: '
            'name, url, path, action, or event.')


class TemplateRefSchema(BaseSchema):
    name = fields.Str(allow_none=True)
    url = fields.Str(allow_none=True)
    path = fields.Str(allow_none=True)
    action = fields.Str(allow_none=True)
    event = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return TemplateRefConfig

    @validates_schema
    def validate_template(self, values):
        validate_template(name=values.get('name'),
                          url=values.get('url'),
                          path=values.get('path'),
                          action=values.get('action'),
                          event=values.get('event'))


class TemplateRefConfig(BaseConfig):
    SCHEMA = TemplateRefSchema
    IDENTIFIER = 'template'
    REDUCED_ATTRIBUTES = ['name', 'url', 'path', 'action', 'event']

    def __init__(self, name=None, url=None, path=None, action=None, event=None):
        validate_template(name=name, url=url, path=path, action=action, event=event)
        self.name = name
        self.url = url
        self.path = path
        self.action = action
        self.event = event



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from hestia.string_utils import strip_spaces
from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.fields import ObjectOrListObject


def get_container_command_args(run_config):
    if not run_config or not run_config.cmd:
        raise ValueError('The specification must contain a command.')

    def sanitize_cmd(cmd):
        cmd = strip_spaces(value=cmd, join=False)
        cmd = [c.strip().strip('\\') for c in cmd if (c and c != '\\')]
        cmd = [c for c in cmd if (c and c != '\\')]
        return ' '.join(cmd)

    return (
        ["/bin/bash", "-c"],
        [' && '.join([sanitize_cmd(cmd) for cmd in run_config.cmd])]
        if isinstance(run_config.cmd, list)
        else [sanitize_cmd(run_config.cmd)]
    )


class RunSchema(BaseSchema):
    cmd = ObjectOrListObject(fields.Str)

    @staticmethod
    def schema_config():
        return RunConfig


class RunConfig(BaseConfig):
    SCHEMA = RunSchema
    IDENTIFIER = 'run'

    def __init__(self, cmd):
        self.cmd = cmd

    def get_container_cmd(self):
        return get_container_command_args(self)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.ops.run import BaseRunConfig, BaseRunSchema, RunSchema


class JobSchema(BaseRunSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('job'))
    run = fields.Nested(RunSchema, allow_none=True)
    backend = fields.Str(allow_none=True)
    framework = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return JobConfig


class JobConfig(BaseRunConfig):
    SCHEMA = JobSchema
    IDENTIFIER = 'job'
    REDUCED_ATTRIBUTES = BaseRunConfig.REDUCED_ATTRIBUTES + ['backend', 'framework', 'run']

    def __init__(self,
                 version=None,
                 kind=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 environment=None,
                 params=None,
                 declarations=None,
                 inputs=None,
                 outputs=None,
                 build=None,
                 backend=None,
                 framework=None,
                 run=None,
                 ):
        super(JobConfig, self).__init__(
            version=version,
            kind=kind,
            logging=logging,
            name=name,
            description=description,
            tags=tags,
            environment=environment,
            params=params,
            declarations=declarations,
            inputs=inputs,
            outputs=outputs,
            build=build
        )

        self.run = run
        self.backend = backend
        self.framework = framework



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import EXCLUDE, fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.fields import ObjectOrListObject, Tensor
from polyaxon_schemas.ml.fields import DType
from polyaxon_schemas.utils import get_value


class BaseLayerSchema(BaseSchema):
    name = fields.Str(allow_none=True)
    trainable = fields.Bool(default=True, missing=True)
    dtype = DType(allow_none=True)
    inbound_nodes = ObjectOrListObject(Tensor, allow_none=True)

    class Meta:
        unknown = EXCLUDE
        ordered = True

    def get_attribute(self, obj, attr, default):
        return get_value(attr, obj, default)


class BaseLayerConfig(BaseConfig):
    REDUCED_ATTRIBUTES = ['name']
    UNKNOWN_BEHAVIOUR = EXCLUDE

    def __init__(self, name=None, trainable=True, dtype='float32', inbound_nodes=None):
        self.name = name
        self.trainable = trainable
        self.dtype = dtype
        self.inbound_nodes = inbound_nodes or []



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields

from polyaxon_schemas.base import BaseConfig, BaseSchema
from polyaxon_schemas.ml.hooks import HookSchema
from polyaxon_schemas.ml.processing.pipelines import PipelineSchema


class TrainSchema(BaseSchema):
    data_pipeline = fields.Nested(PipelineSchema)
    steps = fields.Int(allow_none=True)
    hooks = fields.Nested(HookSchema, many=True, allow_none=True)

    @staticmethod
    def schema_config():
        return TrainConfig


class TrainConfig(BaseConfig):
    SCHEMA = TrainSchema
    IDENTIFIER = 'train'

    def __init__(self, data_pipeline, steps=100, hooks=None):
        self.data_pipeline = data_pipeline
        self.hooks = hooks
        self.steps = steps



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from copy import copy

from hestia.imports import import_string

from polyaxon_schemas.exceptions import PolyaxonConfigurationError
from polyaxon_schemas.ops.operation import BaseOpSchema


class SchemaRegistry(object):
    def __init__(self, base_operations, polyflow_operations):
        self.base_operations = base_operations
        self.polyflow_operations = self._setup_polyflow(polyflow_operations)
        self._schemas = self._setup(base_operations=base_operations,
                                    polyflow_operations=polyflow_operations)

    def _setup(self, base_operations, polyflow_operations):
        schemas = copy(base_operations)
        schemas.update(polyflow_operations)
        return schemas

    def _setup_polyflow(self, polyflow_operations):
        if not polyflow_operations:
            return {}
        ops = {}
        for schema_name, schema_path in polyflow_operations:
            schema = import_string(schema_path)
            if not issubclass(schema, BaseOpSchema):
                raise PolyaxonConfigurationError(
                    'Schema {} at path: {} is not a valid operation.'.format(
                        schema_name, schema_path))
            ops[schema_name] = schema

        return ops



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import ValidationError, fields, validate, validates_schema

from polyaxon_schemas.ops.notebook.backends import NotebookBackend
from polyaxon_schemas.ops.run import BaseRunConfig, BaseRunSchema


def validate_notebook_backend(backend):
    if backend and backend not in NotebookBackend.VALUES:
        raise ValidationError('Notebook backend `{}` not supported'.format(backend))


class NotebookSchema(BaseRunSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal('notebook'))
    backend = fields.Str(allow_none=True, validate=validate.OneOf(NotebookBackend.VALUES))

    @staticmethod
    def schema_config():
        return NotebookConfig

    @validates_schema
    def validate_backend(self, data):
        validate_notebook_backend(data.get('backend'))


class NotebookConfig(BaseRunConfig):
    IDENTIFIER = 'notebook'
    SCHEMA = NotebookSchema
    REDUCED_ATTRIBUTES = BaseRunConfig.REDUCED_ATTRIBUTES + ['backend']

    def __init__(self,
                 version=None,
                 kind=None,
                 logging=None,
                 name=None,
                 description=None,
                 tags=None,
                 environment=None,
                 params=None,
                 declarations=None,
                 inputs=None,
                 outputs=None,
                 build=None,
                 backend=None,
                 ):
        super(NotebookConfig, self).__init__(
            version=version,
            kind=kind,
            logging=logging,
            name=name,
            description=description,
            tags=tags,
            environment=environment,
            params=params,
            declarations=declarations,
            inputs=inputs,
            outputs=outputs,
            build=build
        )
        validate_notebook_backend(backend)
        self.backend = backend



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from marshmallow import fields, validate

from polyaxon_schemas.base import BaseConfig, BaseMultiSchema, BaseSchema
from polyaxon_schemas.ml.fields import DType

# pylint:disable=too-many-lines


class ZerosInitializerSchema(BaseSchema):
    dtype = DType(allow_none=True)

    @staticmethod
    def schema_config():
        return ZerosInitializerConfig


class ZerosInitializerConfig(BaseConfig):
    """Initializer that generates tensors initialized to 0.

    Args:
        dtype: The data type.

    Returns:
        An initializer.

    Polyaxonfile usage:

    Using the default values

    ```yaml
    Zeros:
    ```

    Using custom values

    ```yaml
    Zeros:
      dtype: int16
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: Zeros
    ```
    """
    IDENTIFIER = 'Zeros'
    SCHEMA = ZerosInitializerSchema

    def __init__(self, dtype='float32'):
        self.dtype = dtype


class OnesInitializerSchema(BaseSchema):
    dtype = DType(allow_none=True)

    @staticmethod
    def schema_config():
        return OnesInitializerConfig


class OnesInitializerConfig(BaseConfig):
    """Initializer that generates tensors initialized to 1.

    Args:
        dtype: The data type.

    Returns:
        An initializer.

    Polyaxonfile usage:

    Using the default values

    ```yaml
    Ones:
    ```

    Using custom values

    ```yaml
    Ones:
      dtype: int16
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: Ones
    ```
    """
    IDENTIFIER = 'Ones'
    SCHEMA = OnesInitializerSchema

    def __init__(self, dtype='float32'):
        self.dtype = dtype


class ConstantInitializerSchema(BaseSchema):
    value = fields.Int(allow_none=True)
    dtype = DType(allow_none=True)

    @staticmethod
    def schema_config():
        return ConstantInitializerConfig


class ConstantInitializerConfig(BaseConfig):
    """Initializer that generates tensors with constant values.

    The resulting tensor is populated with values of type `dtype`, as
    specified by arguments `value` following the desired `shape` of the
    new tensor (see examples below).

    The argument `value` can be a constant value, or a list of values of type
    `dtype`. If `value` is a list, then the length of the list must be less
    than or equal to the number of elements implied by the desired shape of the
    tensor. In the case where the total number of elements in `value` is less
    than the number of elements required by the tensor shape, the last element
    in `value` will be used to fill the remaining entries. If the total number of
    elements in `value` is greater than the number of elements required by the
    tensor shape, the initializer will raise a `ValueError`.

    Args:
        value: A Python scalar, list of values, or a N-dimensional numpy array. All
            elements of the initialized variable will be set to the corresponding
            value in the `value` argument.
        dtype: The data type.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    Constant:
      value: 3
      dtype: int16
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        Constant:
          value: 3
    ```
    """
    IDENTIFIER = 'Constant'
    SCHEMA = ConstantInitializerSchema

    def __init__(self, value=0, dtype='float32'):
        self.dtype = dtype
        self.value = value


class UniformInitializerSchema(BaseSchema):
    minval = fields.Number(allow_none=True)
    maxval = fields.Number(allow_none=True)
    dtype = DType(allow_none=True)
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return UniformInitializerConfig


class UniformInitializerConfig(BaseConfig):
    """Initializer that generates tensors with a uniform distribution.

    Args:
        minval: A python scalar or a scalar tensor. Lower bound of the range
            of random values to generate.
        maxval: A python scalar or a scalar tensor. Upper bound of the range
            of random values to generate.  Defaults to 1 for float types.
        seed: A Python integer. Used to create random seeds. See
            @{tf.set_random_seed} for behavior.
        dtype: The data type.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    Uniform:
      minval: 1
      maxval: 2
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: Uniform
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        Uniform:
          minval: 1
    ```
    """
    IDENTIFIER = 'Uniform'
    SCHEMA = UniformInitializerSchema

    def __init__(self, minval=0, maxval=None, seed=None, dtype='float32'):
        self.seed = seed
        self.dtype = dtype
        self.minval = minval
        self.maxval = maxval


class NormalInitializerSchema(BaseSchema):
    mean = fields.Number(allow_none=True)
    stddev = fields.Number(allow_none=True)
    dtype = DType(allow_none=True)
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return NormalInitializerConfig


class NormalInitializerConfig(BaseConfig):
    """Initializer that generates tensors with a normal distribution.

    Args:
        mean: a python scalar or a scalar tensor. Mean of the random values to generate.
        stddev: a python scalar or a scalar tensor. Standard deviation of the
            random values to generate.
        seed: A Python integer. Used to create random seeds. See
            @{tf.set_random_seed} for behavior.
        dtype: The data type. Only floating point types are supported.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    Normal:
      mean: 0.5
      stddev: 1.
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: Normal
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        Normal:
          mean: 1.
    ```
    """
    IDENTIFIER = 'Normal'
    SCHEMA = NormalInitializerSchema

    def __init__(self, mean=0., stddev=1., seed=None, dtype='float32'):
        self.seed = seed
        self.dtype = dtype
        self.mean = mean
        self.stddev = stddev


class TruncatedNormalInitializerSchema(BaseSchema):
    mean = fields.Number(allow_none=True)
    stddev = fields.Number(allow_none=True)

    @staticmethod
    def schema_config():
        return TruncatedNormalInitializerConfig


class TruncatedNormalInitializerConfig(BaseConfig):
    """Initializer that generates a truncated normal distribution.

    These values are similar to values from a `random_normal_initializer`
    except that values more than two standard deviations from the mean
    are discarded and re-drawn. This is the recommended initializer for
    neural network weights and filters.

    Args:
        mean: a python scalar or a scalar tensor. Mean of the random values to generate.
        stddev: a python scalar or a scalar tensor. Standard deviation of the
            random values to generate.
        seed: A Python integer. Used to create random seeds. See
            @{tf.set_random_seed} for behavior.
        dtype: The data type. Only floating point types are supported.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    TruncatedNormal:
      mean: 0.5
      stddev: 1.
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: TruncatedNormal
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        TruncatedNormal:
          mean: 1.
    ```
    """
    IDENTIFIER = 'TruncatedNormal'
    SCHEMA = TruncatedNormalInitializerSchema

    def __init__(self, mean=0., stddev=1., seed=None, dtype='float32'):
        self.seed = seed
        self.dtype = dtype
        self.mean = mean
        self.stddev = stddev


class VarianceScalingInitializerSchema(BaseSchema):
    scale = fields.Float(allow_none=True)
    mode = fields.Str(allow_none=True, validate=validate.OneOf(['fan_in', 'fan_out', 'fan_avg']))
    distribution = fields.Str(allow_none=True)
    dtype = DType(allow_none=True)

    @staticmethod
    def schema_config():
        return VarianceScalingInitializerConfig


class VarianceScalingInitializerConfig(BaseConfig):
    """Initializer capable of adapting its scale to the shape of weights tensors.

    With `distribution="normal"`, samples are drawn from a truncated normal
    distribution centered on zero, with `stddev = sqrt(scale / n)`
    where n is:
      - number of input units in the weight tensor, if mode = "fan_in"
      - number of output units, if mode = "fan_out"
      - average of the numbers of input and output units, if mode = "fan_avg"

    With `distribution="uniform"`, samples are drawn from a uniform distribution
    within [-limit, limit], with `limit = sqrt(3 * scale / n)`.

    Args:
        scale: Scaling factor (positive float).
        mode: One of "fan_in", "fan_out", "fan_avg".
        distribution: Random distribution to use. One of "normal", "uniform".
        seed: A Python integer. Used to create random seeds. See
            @{tf.set_random_seed} for behavior.
        dtype: The data type. Only floating point types are supported.

    Raises:
        ValueError: In case of an invalid value for the "scale", mode" or "distribution" arguments.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    VarianceScaling:
      scale: 0.5
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: VarianceScaling
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        VarianceScaling:
          scale: 1.
          mode: fan_out
    ```
    """
    IDENTIFIER = 'VarianceScaling'
    SCHEMA = VarianceScalingInitializerSchema

    def __init__(self, scale=1., mode='fan_in', distribution="normal", dtype='float32'):
        self.scale = scale
        self.mode = mode
        self.distribution = distribution
        self.dtype = dtype


class IdentityInitializerSchema(BaseSchema):
    gain = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return IdentityInitializerConfig


class IdentityInitializerConfig(BaseConfig):
    """Initializer that generates the identity matrix.

    Only use for 2D matrices.

    Args:
        gain: Multiplicative factor to apply to the identity matrix.
        dtype: The type of the output.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    Identity:
      gain: 0.5
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: Identity
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        Identity:
          gain: 0.5
    ```
    """
    IDENTIFIER = 'Identity'
    SCHEMA = IdentityInitializerSchema

    def __init__(self, gain=1.):
        self.gain = gain


class OrthogonalInitializerSchema(BaseSchema):
    mean = fields.Number(allow_none=True)
    stddev = fields.Number(allow_none=True)
    gain = fields.Float(allow_none=True)

    @staticmethod
    def schema_config():
        return OrthogonalInitializerConfig


class OrthogonalInitializerConfig(BaseConfig):
    """Initializer that generates an orthogonal matrix.

    If the shape of the tensor to initialize is two-dimensional, it is initialized
    with an orthogonal matrix obtained from the QR decomposition of a matrix of
    uniform random numbers. If the matrix has fewer rows than columns then the
    output will have orthogonal rows. Otherwise, the output will have orthogonal
    columns.

    If the shape of the tensor to initialize is more than two-dimensional,
    a matrix of shape `(shape[0] * ... * shape[n - 2], shape[n - 1])`
    is initialized, where `n` is the length of the shape vector.
    The matrix is subsequently reshaped to give a tensor of the desired shape.

    Args:
        gain: multiplicative factor to apply to the orthogonal matrix
        dtype: The type of the output.
        seed: A Python integer. Used to create random seeds. See
            @{tf.set_random_seed} for behavior.

    Returns:
        An initializer.

    Polyaxonfile usage:

    ```yaml
    Orthogonal:
      gain: 0.5
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: Orthogonal
    ```

    or

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer:
        Orthogonal:
          gain: 0.5
    ```
    """
    IDENTIFIER = 'Orthogonal'
    SCHEMA = OrthogonalInitializerSchema

    def __init__(self, gain=1., seed=None, dtype='float32'):
        self.seed = seed
        self.dtype = dtype
        self.gain = gain


class GlorotUniformInitializerSchema(BaseSchema):
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return GlorotUniformInitializerConfig


class GlorotUniformInitializerConfig(BaseConfig):
    """Glorot uniform initializer, also called Xavier uniform initializer.

    It draws samples from a uniform distribution within [-limit, limit]
    where `limit` is `sqrt(6 / (fan_in + fan_out))`
    where `fan_in` is the number of input units in the weight tensor
    and `fan_out` is the number of output units in the weight tensor.

    Args:
        seed: A Python integer. Used to seed the random generator.

    Returns:
        An initializer.

    References:
        Glorot & Bengio, AISTATS 2010
        http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf

    Polyaxonfile usage:

    Using the default values

    ```yaml
    GlorotUniform:
    ```

    Using custom values

    ```yaml
    GlorotUniform:
      seed: 10
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: GlorotUniform
    ```
    """
    IDENTIFIER = 'GlorotUniform'
    SCHEMA = GlorotUniformInitializerSchema

    def __init__(self, seed=None):
        self.seed = seed


class GlorotNormalInitializerSchema(BaseSchema):
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return GlorotNormalInitializerConfig


class GlorotNormalInitializerConfig(BaseConfig):
    """Glorot normal initializer, also called Xavier normal initializer.

    It draws samples from a truncated normal distribution centered on 0
    with `stddev = sqrt(2 / (fan_in + fan_out))`
    where `fan_in` is the number of input units in the weight tensor
    and `fan_out` is the number of output units in the weight tensor.

    Args:
        seed: A Python integer. Used to seed the random generator.

    Returns:
        An initializer.

    References:
        Glorot & Bengio, AISTATS 2010
        http://jmlr.org/proceedings/papers/v9/glorot10a/glorot10a.pdf

    Polyaxonfile usage:

    Using the default values

    ```yaml
    GlorotNormal:
    ```

    Using custom values

    ```yaml
    GlorotNormal:
      seed: 10
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: GlorotNormal
    ```
    """
    IDENTIFIER = 'GlorotNormal'
    SCHEMA = GlorotNormalInitializerSchema

    def __init__(self, seed=None):
        self.seed = seed


class HeUniformInitializerSchema(BaseSchema):
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return HeUniformInitializerConfig


class HeUniformInitializerConfig(BaseConfig):
    """He uniform variance scaling initializer.

    It draws samples from a uniform distribution within [-limit, limit]
    where `limit` is `sqrt(6 / fan_in)`
    where `fan_in` is the number of input units in the weight tensor.

    Args:
        seed: A Python integer. Used to seed the random generator.

    Returns:
        An initializer.

    References:
        He et al., http://arxiv.org/abs/1502.01852

    Polyaxonfile usage:

    Using the default values

    ```yaml
    HeUniform:
    ```

    Using custom values

    ```yaml
    HeUniform:
      seed: 10
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: HeUniform
    ```
    """
    IDENTIFIER = 'HeUniform'
    SCHEMA = HeUniformInitializerSchema

    def __init__(self, seed=None):
        self.seed = seed


class HeNormalInitializerSchema(BaseSchema):
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return HeNormalInitializerConfig


class HeNormalInitializerConfig(BaseConfig):
    """He normal initializer.

    It draws samples from a truncated normal distribution centered on 0
    with `stddev = sqrt(2 / fan_in)`
    where `fan_in` is the number of input units in the weight tensor.

    Args:
        seed: A Python integer. Used to seed the random generator.

    Returns:
        An initializer.

    References:
        He et al., http://arxiv.org/abs/1502.01852

    Polyaxonfile usage:

    Using the default values

    ```yaml
    HeNormal:
    ```

    Using custom values

    ```yaml
    HeNormal:
      seed: 10
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: HeNormal
    ```
    """
    IDENTIFIER = 'HeNormal'
    SCHEMA = HeNormalInitializerSchema

    def __init__(self, seed=None):
        self.seed = seed


class LecunUniformInitializerSchema(BaseSchema):
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return LecunUniformInitializerConfig


class LecunUniformInitializerConfig(BaseConfig):
    """LeCun uniform initializer.

    It draws samples from a uniform distribution within [-limit, limit]
    where `limit` is `sqrt(3 / fan_in)`
    where `fan_in` is the number of input units in the weight tensor.

    Args:
        seed: A Python integer. Used to seed the random generator.

    Returns:
        An initializer.

    References:
        LeCun 98, Efficient Backprop,
        http://yann.lecun.com/exdb/publis/pdf/lecun-98b.pdf

    Polyaxonfile usage:

    Using the default values

    ```yaml
    LecunUniform:
    ```

    Using custom values

    ```yaml
    LecunUniform:
      seed: 10
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: LecunUniform
    ```
    """
    IDENTIFIER = 'LecunUniform'
    SCHEMA = LecunUniformInitializerSchema

    def __init__(self, seed=None):
        self.seed = seed


class LecunNormalInitializerSchema(BaseSchema):
    seed = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return LecunNormalInitializerConfig


class LecunNormalInitializerConfig(BaseConfig):
    """LeCun normal initializer.

    It draws samples from a truncated normal distribution centered on 0
    with `stddev = sqrt(1 / fan_in)`
    where `fan_in` is the number of input units in the weight tensor.

    Args:
        seed: A Python integer. Used to seed the random generator.

    Returns:
        An initializer.

    References:
        - [Self-Normalizing Neural Networks](https://arxiv.org/abs/1706.02515)
        - [Efficient
        Backprop](http://yann.lecun.com/exdb/publis/pdf/lecun-98b.pdf)

    Polyaxonfile usage:

    Using the default values

    ```yaml
    LecunNormal:
    ```

    Using custom values

    ```yaml
    LecunNormal:
      seed: 10
    ```

    Example with layer

    ```yaml
    Conv2D:
      filters: 10
      kernel_size: 8
      kernel_initializer: LecunNormal
    ```
    """
    IDENTIFIER = 'LecunNormal'
    SCHEMA = LecunNormalInitializerSchema

    def __init__(self, seed=None):
        self.seed = seed


class InitializerSchema(BaseMultiSchema):
    __multi_schema_name__ = 'initializer'
    __configs__ = {
        ZerosInitializerConfig.IDENTIFIER: ZerosInitializerConfig,
        OnesInitializerConfig.IDENTIFIER: OnesInitializerConfig,
        ConstantInitializerConfig.IDENTIFIER: ConstantInitializerConfig,
        UniformInitializerConfig.IDENTIFIER: UniformInitializerConfig,
        NormalInitializerConfig.IDENTIFIER: NormalInitializerConfig,
        TruncatedNormalInitializerConfig.IDENTIFIER: TruncatedNormalInitializerConfig,
        VarianceScalingInitializerConfig.IDENTIFIER: VarianceScalingInitializerConfig,
        IdentityInitializerConfig.IDENTIFIER: IdentityInitializerConfig,
        OrthogonalInitializerConfig.IDENTIFIER: OrthogonalInitializerConfig,
        GlorotUniformInitializerConfig.IDENTIFIER: GlorotUniformInitializerConfig,
        GlorotNormalInitializerConfig.IDENTIFIER: GlorotNormalInitializerConfig,
        HeUniformInitializerConfig.IDENTIFIER: HeUniformInitializerConfig,
        HeNormalInitializerConfig.IDENTIFIER: HeNormalInitializerConfig,
        LecunUniformInitializerConfig.IDENTIFIER: LecunUniformInitializerConfig,
        LecunNormalInitializerConfig.IDENTIFIER: LecunNormalInitializerConfig,
    }
    __support_snake_case__ = True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from marshmallow import ValidationError, fields


class Tensor(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, six.string_types):
            return [value, 0, 0]
        if isinstance(value, list) and len(value) == 3:
            condition = (isinstance(value[0], str) and
                         isinstance(value[1], int) and
                         isinstance(value[1], int))
            if condition:
                return value
        raise ValidationError("This field expects a str or a list of [str, int, int].")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from unittest import TestCase

from marshmallow import ValidationError
from tests.utils import assert_equal_dict

from polyaxon_schemas.ops.io import IOTypes
from polyaxon_schemas.ops.io.io import IOConfig


class TestIOConfigs(TestCase):
    def test_wrong_io_config(self):
        # No name
        with self.assertRaises(TypeError):
            IOConfig.from_dict({})

    def test_unsupported_io_config_type(self):
        with self.assertRaises(ValidationError):
            IOConfig.from_dict({
                'name': 'input1',
                'type': 'something'
            })

    def test_wrong_io_config_default(self):
        with self.assertRaises(ValidationError):
            IOConfig.from_dict({
                'name': 'input1',
                'type': IOTypes.FLOAT,
                'default': 'foo'
            })

        with self.assertRaises(ValidationError):
            IOConfig.from_dict({
                'name': 'input1',
                'type': IOTypes.GCS_PATH,
                'default': 234
            })

    def test_wrong_io_config_flag(self):
        with self.assertRaises(ValidationError):
            IOConfig.from_dict({
                'name': 'input1',
                'type': IOTypes.S3_PATH,
                'is_flag': True
            })

        with self.assertRaises(ValidationError):
            IOConfig.from_dict({
                'name': 'input1',
                'type': IOTypes.FLOAT,
                'is_flag': True
            })

    def test_io_config_optionals(self):
        config_dict = {
            'name': 'input1',
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_io_config_desc(self):
        # test desc
        config_dict = {
            'name': 'input1',
            'description': 'some text'
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_io_config_types(self):
        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.INT
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.S3_PATH
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_io_config_default(self):
        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.BOOL,
            'is_optional': True,
            'default': True,
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.FLOAT,
            'is_optional': True,
            'default': 3.4,
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_io_config_default_and_required(self):
        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.BOOL,
            'default': True,
            'is_optional': True
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.STR,
            'default': 'foo'
        }
        with self.assertRaises(ValidationError):
            IOConfig.from_dict(config_dict)

    def test_io_config_required(self):
        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': 'float',
            'is_optional': False,
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_io_config_flag(self):
        config_dict = {
            'name': 'input1',
            'description': 'some text',
            'type': IOTypes.BOOL,
            'is_flag': True,
        }
        config = IOConfig.from_dict(config_dict)
        assert_equal_dict(config.to_dict(), config_dict)

    def test_value_non_typed_input(self):
        config_dict = {
            'name': 'input1'
        }
        config = IOConfig.from_dict(config_dict)
        assert config.validate_value('foo') == 'foo'
        assert config.validate_value(1) == 1
        assert config.validate_value(True) is True

    def test_value_typed_input(self):
        config_dict = {
            'name': 'input1',
            'type': IOTypes.BOOL
        }
        config = IOConfig.from_dict(config_dict)
        with self.assertRaises(ValidationError):
            config.validate_value('foo')
        with self.assertRaises(ValidationError):
            config.validate_value(1)
        with self.assertRaises(ValidationError):
            config.validate_value(None)

        assert config.validate_value(True) is True

    def test_value_typed_input_with_default(self):
        config_dict = {
            'name': 'input1',
            'type': IOTypes.INT,
            'default': 12,
            'is_optional': True,
        }
        config = IOConfig.from_dict(config_dict)
        with self.assertRaises(ValidationError):
            config.validate_value('foo')

        assert config.validate_value(1) == 1
        assert config.validate_value(0) == 0
        assert config.validate_value(-1) == -1
        assert config.validate_value(None) == 12



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import six

from collections import Mapping, OrderedDict

from hestia.humanize import humanize_timesince
from hestia.tz_utils import get_time_zone
from hestia.units import to_percentage, to_unit_memory
from marshmallow import RAISE, Schema, ValidationError, post_dump, post_load
from marshmallow.utils import EXCLUDE, utc

from polyaxon_schemas.exceptions import PolyaxonSchemaError
from polyaxon_schemas.utils import to_camel_case


class BaseSchema(Schema):
    """Base schema."""

    class Meta:
        unknown = RAISE
        ordered = True

    @post_load
    def make(self, data):
        return self.schema_config()(**data)

    @post_dump
    def unmake(self, data):
        return self.schema_config().remove_reduced_attrs(data)

    @staticmethod
    def schema_config():
        raise NotImplementedError()


class BaseConfig(object):
    """Base for config classes."""

    SCHEMA = None
    IDENTIFIER = None
    REDUCED_ATTRIBUTES = []  # Attribute to remove in the reduced form if they are null.
    DEFAULT_INCLUDE_ATTRIBUTES = []
    DEFAULT_EXCLUDE_ATTRIBUTES = []
    DATETIME_ATTRIBUTES = []
    MEM_SIZE_ATTRIBUTES = []
    PERCENT_ATTRIBUTES = []
    ROUNDING = 2
    UNKNOWN_BEHAVIOUR = RAISE

    def to_light_dict(self,
                      humanize_values=False,
                      include_attrs=None,
                      exclude_attrs=None,
                      unknown=None):
        unknown = unknown or self.UNKNOWN_BEHAVIOUR
        obj_dict = self.to_dict(humanize_values=humanize_values, unknown=unknown)
        if all([include_attrs, exclude_attrs]):
            raise PolyaxonSchemaError(
                'Only one value `include_attrs` or `exclude_attrs` is allowed.')
        if not any([include_attrs, exclude_attrs]):  # Use Default setup attrs
            include_attrs = self.DEFAULT_INCLUDE_ATTRIBUTES
            exclude_attrs = self.DEFAULT_EXCLUDE_ATTRIBUTES

        if include_attrs:
            exclude_attrs = set(six.iterkeys(obj_dict)) - set(include_attrs)
        for attr in exclude_attrs:
            obj_dict.pop(attr, None)

        return obj_dict

    def to_dict(self, humanize_values=False, unknown=None):
        unknown = unknown or self.UNKNOWN_BEHAVIOUR
        return self.obj_to_dict(self, humanize_values=humanize_values, unknown=unknown)

    def to_schema(self):
        return self.obj_to_schema(self)

    @classmethod
    def humanize_attrs(cls, obj):
        humanized_attrs = {}
        for attr in cls.DATETIME_ATTRIBUTES:
            humanized_attrs[attr] = humanize_timesince(getattr(obj, attr))
        for attr in cls.PERCENT_ATTRIBUTES:
            humanized_attrs[attr] = to_percentage(getattr(obj, attr), cls.ROUNDING)
        for attr in cls.MEM_SIZE_ATTRIBUTES:
            humanized_attrs[attr] = to_unit_memory(getattr(obj, attr))
        return humanized_attrs

    @classmethod
    def obj_to_dict(cls, obj, humanize_values=False, unknown=None):
        unknown = unknown or cls.UNKNOWN_BEHAVIOUR
        humanized_attrs = cls.humanize_attrs(obj) if humanize_values else {}
        data_dict = cls.SCHEMA(unknown=unknown).dump(obj)  # pylint: disable=not-callable

        for k, v in six.iteritems(humanized_attrs):
            data_dict[k] = v
        return data_dict

    @classmethod
    def remove_reduced_attrs(cls, data):
        obj_dict = OrderedDict((key, value) for (key, value) in six.iteritems(data))
        for attr in cls.REDUCED_ATTRIBUTES:
            if obj_dict[attr] is None:
                del obj_dict[attr]

        return obj_dict

    @classmethod
    def obj_to_schema(cls, obj):
        return {cls.IDENTIFIER: cls.obj_to_dict(obj)}

    @classmethod
    def from_dict(cls, value, unknown=None):
        unknown = unknown or cls.UNKNOWN_BEHAVIOUR
        return cls.SCHEMA(unknown=unknown).load(value)  # pylint: disable=not-callable

    @staticmethod
    def localize_date(dt):
        if not dt:
            return dt
        if not dt.tzinfo:
            dt = utc.localize(dt)
        return dt.astimezone(get_time_zone())

    @classmethod
    def to_jsonschema(cls):
        from marshmallow_jsonschema import JSONSchema  # pylint:disable=import-error

        return JSONSchema().dump(cls.SCHEMA())  # pylint:disable=not-callable


class BaseOneOfSchema(Schema):
    """
    Code taken and adapted from: marshmallow-oneofschema

    This is a special kind of schema that actually multiplexes other schemas
    based on object type. When serializing values, it uses get_obj_type() method
    to get object type name. Then it uses `SCHEMAS` name-to-Schema mapping
    to get schema for that particular object type, serializes object using that
    schema and adds an extra "type" field with name of object type.
    Deserialization is reverse.
    """
    TYPE_FIELD = 'type'
    TYPE_FIELD_REMOVE = True
    SCHEMAS = {}

    class Meta:
        unknown = RAISE

    def get_obj_type(self, obj):
        """Returns name of object schema"""
        return obj.IDENTIFIER

    def dump(self, obj, many=None):
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if not many:
            result_data = self._dump(obj)
        else:
            for idx, o in enumerate(obj):
                try:
                    result = self._dump(o)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.messages
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=obj, valid_data=result)
            raise exc

    def _dump(self, obj):
        obj_type = self.get_obj_type(obj)
        if not obj_type:
            return None, {'_schema': 'Unknown object class: %s' % obj.__class__.__name__}

        type_schema = self.SCHEMAS.get(obj_type)
        if not type_schema:
            return None, {'_schema': 'Unsupported object type: %s' % obj_type}

        schema = (
            type_schema if isinstance(type_schema, Schema)
            else type_schema()
        )

        schema.context.update(getattr(self, 'context', {}))

        result = schema.dump(obj, many=False)
        if result is not None:
            result[self.TYPE_FIELD] = obj_type
        return result

    def load(self, data, many=None, partial=None, unknown=None):
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if partial is None:
            partial = self.partial
        if not many:
            try:
                result_data = self._load(data, partial=partial, unknown=unknown)
            except ValidationError as error:
                result_errors[0] = error.messages
                result_data.append(error.valid_data)
        else:
            for idx, item in enumerate(data):
                try:
                    result = self._load(item, partial=partial)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.messages
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=data, valid_data=result)
            raise exc

    def _load(self, data, partial=None, unknown=None):
        if not isinstance(data, dict):
            raise ValidationError({'_schema': 'Invalid data type: %s' % data})

        data = dict(data)
        unknown = unknown or self.unknown

        data_type = data.get(self.TYPE_FIELD)
        if self.TYPE_FIELD in data and self.TYPE_FIELD_REMOVE:
            data.pop(self.TYPE_FIELD)

        if not data_type:
            raise ValidationError({
                self.TYPE_FIELD: ['Missing data for required field.']
            })

        try:
            type_schema = self.SCHEMAS.get(data_type)
        except TypeError:
            # data_type could be unhashable
            raise ValidationError({
                self.TYPE_FIELD: ['Invalid value: %s' % data_type]
            })
        if not type_schema:
            raise ValidationError({
                self.TYPE_FIELD: ['Unsupported value: %s' % data_type],
            })

        schema = (
            type_schema if isinstance(type_schema, Schema) else type_schema()
        )

        schema.context.update(getattr(self, 'context', {}))

        return schema.load(data, many=False, partial=partial, unknown=unknown)

    def validate(self, data, many=None, partial=None):
        try:
            self.load(data, many=many, partial=partial)
        except ValidationError as ve:
            return ve.messages
        return {}


class BaseMultiSchema(Schema):
    __multi_schema_name__ = None
    __configs__ = None
    # to support snake case identifier, e.g. glorot_uniform and GlorotUniform
    __support_snake_case__ = False

    class Meta:
        unknown = EXCLUDE

    @post_dump(pass_original=True, pass_many=True)
    def handle_multi_schema_dump(self, data, pass_many, original):
        def handle_item(item):
            if hasattr(item, 'get_config'):
                return self.__configs__[item.__class__.__name__].obj_to_schema(item)
            return item.to_schema()

        if pass_many:
            return [handle_item(item) for item in original]

        return handle_item(original)

    @post_load(pass_original=True, pass_many=True)
    def handle_multi_schema_load(self, data, pass_many, original):
        def make(key, val=None):
            key = to_camel_case(key) if self.__support_snake_case__ else key
            try:
                return (self.__configs__[key].from_dict(val, unknown=EXCLUDE) if val else
                        self.__configs__[key]())
            except KeyError:
                raise ValidationError("`{}` is not a valid value for schema `{}`".format(
                    key, self.__multi_schema_name__))

        def handle_item(item):
            if isinstance(item, six.string_types):
                return make(item)

            if isinstance(item, Mapping):
                if 'class_name' in item:
                    return make(item['class_name'], item['config'])
                if 'model_type' in item:
                    return make(item.pop('model_type'), item)
                assert len(item) == 1
                key, val = list(six.iteritems(item))[0]
                return make(key, val)

        if pass_many:
            return [handle_item(item) for item in original]

        return handle_item(original)


NAME_REGEX = r'^[-a-zA-Z0-9_]+\Z'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import numpy as np

from unittest import TestCase

from marshmallow.exceptions import ValidationError

from polyaxon_schemas.ops.group.matrix import MatrixConfig


class TestMatrixConfigs(TestCase):
    def test_matrix_requires_at_least_one_value(self):
        config_dict = {}
        with self.assertRaises(ValidationError):
            MatrixConfig.from_dict(config_dict)

    def test_matrix_accept_only_one_option(self):
        config_dict = {
            'values': [1, 2, 3],
            'linspace': '1:2:1'
        }
        with self.assertRaises(ValidationError):
            MatrixConfig.from_dict(config_dict)

    def test_matrix_values_option(self):
        config_dict = {
            'values': [1, 2, 3],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.to_numpy() == config_dict['values']
        assert config.sample() in [1, 2, 3]
        assert config.length == 3
        assert config.is_categorical is False
        assert config.is_distribution is False
        assert config.is_range is False
        assert config.is_uniform is False
        assert config.is_discrete is True
        assert config.is_continuous is False
        assert config.min == 1
        assert config.max == 3

        config_dict['values'] = ['ok', 'nook']
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.to_numpy() == config_dict['values']
        assert config.sample() in ['ok', 'nook']
        assert config.length == 2
        assert config.is_categorical is True
        assert config.is_distribution is False
        assert config.is_range is False
        assert config.is_uniform is False
        assert config.is_discrete is True
        assert config.is_continuous is False
        assert config.min is None
        assert config.max is None

        config_dict['values'] = [[1, 2], [2, 4]]
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        assert config.to_numpy() == config_dict['values']
        assert config.sample() in [[1, 2], [2, 4]]
        assert config.length == 2
        assert config.is_categorical is True
        assert config.is_distribution is False
        assert config.is_range is False
        assert config.is_uniform is False
        assert config.is_discrete is True
        assert config.is_continuous is False
        assert config.min is None
        assert config.max is None

    def test_matrix_pvalues_option(self):
        config_dict = {
            'pvalues': [(1, 0.1), (2, 0.3), (3, 6)],
        }
        with self.assertRaises(ValidationError):
            MatrixConfig.from_dict(config_dict)

        config_dict = {
            'pvalues': [(1, 0.1), (2, 0.3), (3, 0.8)],
        }
        with self.assertRaises(ValidationError):
            MatrixConfig.from_dict(config_dict)

        config_dict = {
            'pvalues': [(1, 0.1), (2, 0.3), (3, -0.6)],
        }
        with self.assertRaises(ValidationError):
            MatrixConfig.from_dict(config_dict)

        config_dict['pvalues'] = ['ok', 'nook']
        with self.assertRaises(ValidationError):
            MatrixConfig.from_dict(config_dict)

        # Pass for correct config
        config_dict = {
            'pvalues': [(1, 0.1), (2, 0.1), (3, 0.8)],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
        with self.assertRaises(ValidationError):
            config.to_numpy()
        assert config.sample() in [1, 2, 3]
        assert config.length == 3
        assert config.is_categorical is False
        assert config.is_distribution is True
        assert config.is_range is False
        assert config.is_uniform is False
        assert config.is_discrete is True
        assert config.is_continuous is False
        assert config.min is None
        assert config.max is None

    def test_matrix_range_option(self):
        def assert_equal(config, v1, v2, v3):
            result = {'start': v1, 'stop': v2, 'step': v3}
            assert config.to_dict()['range'] == result
            np.testing.assert_array_equal(config.to_numpy(), np.arange(**result))
            assert config.length == len(np.arange(**result))
            assert config.sample() in np.arange(**result)
            assert config.is_categorical is False
            assert config.is_distribution is False
            assert config.is_range is True
            assert config.is_uniform is False
            assert config.is_discrete is True
            assert config.is_continuous is False
            assert config.min == v1
            assert config.max == v2

        # as list
        config_dict = {
            'range': [1, 2, 3],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['range'])

        # as string
        config_dict['range'] = '0:10:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 10, 1)

        # as dict
        config_dict['range'] = {'start': 1.2, 'stop': 1.8, 'step': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_linspace_option(self):
        def assert_equal(config, v1, v2, v3):
            result = {'start': v1, 'stop': v2, 'num': v3}
            assert config.to_dict()['linspace'] == result
            np.testing.assert_array_equal(config.to_numpy(), np.linspace(**result))
            assert config.length == len(np.linspace(**result))
            assert config.sample() in np.linspace(**result)
            assert config.is_categorical is False
            assert config.is_distribution is False
            assert config.is_range is True
            assert config.is_uniform is False
            assert config.is_discrete is True
            assert config.is_continuous is False
            assert config.min == v1
            assert config.max == v2

        # as list
        config_dict = {
            'linspace': [1, 2, 3],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['linspace'])

        # as string
        config_dict['linspace'] = '0:10:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 10, 1)

        # as dict
        config_dict['linspace'] = {'start': 1.2, 'stop': 1.8, 'num': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_geomspace_option(self):
        def assert_equal(config, v1, v2, v3):
            result = {'start': v1, 'stop': v2, 'num': v3}
            assert config.to_dict()['geomspace'] == result
            np.testing.assert_array_equal(config.to_numpy(), np.geomspace(**result))
            assert config.length == len(np.geomspace(**result))
            assert config.sample() in np.geomspace(**result)
            assert config.is_categorical is False
            assert config.is_distribution is False
            assert config.is_range is True
            assert config.is_uniform is False
            assert config.is_discrete is True
            assert config.is_continuous is False
            assert config.min == v1
            assert config.max == v2

        # as list
        config_dict = {
            'geomspace': [1, 2, 3],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['geomspace'])

        # as string
        config_dict['geomspace'] = '1:10:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 1, 10, 1)

        # as dict
        config_dict['geomspace'] = {'start': 1.2, 'stop': 1.8, 'num': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_logspace_option(self):
        def assert_equal(config, v1, v2, v3, v4=None):
            result = {'start': v1, 'stop': v2, 'num': v3}
            if v4:
                result['base'] = v4

            assert config.to_dict()['logspace'] == result
            np.testing.assert_array_equal(config.to_numpy(), np.logspace(**result))
            assert config.length == len(np.logspace(**result))
            assert config.sample() in np.logspace(**result)
            assert config.is_categorical is False
            assert config.is_distribution is False
            assert config.is_range is True
            assert config.is_uniform is False
            assert config.is_discrete is True
            assert config.is_continuous is False
            assert config.min == v1
            assert config.max == v2

        # as list
        config_dict = {
            'logspace': [1, 2, 3],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['logspace'])

        # with base
        config_dict['logspace'] = [1, 2, 3, 2]
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['logspace'])

        # as string
        config_dict['logspace'] = '0:10:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 10, 1)

        # with base
        config_dict['logspace'] = '0:10:1:2'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 10, 1, 2)

        # as dict
        config_dict['logspace'] = {'start': 1.2, 'stop': 1.8, 'num': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

        # with base
        config_dict['logspace'] = {'start': 1.2, 'stop': 1.8, 'num': 0.1, 'base': 2}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_uniform_option(self):
        def assert_equal(config, v1, v2, v3=None):
            result = {'low': v1, 'high': v2}
            if v3:
                result['size'] = v3
            assert config.to_dict()['uniform'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert v1 <= config.sample() <= v2
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is True
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min == v1
            assert config.max == v2

        # as list
        config_dict = {
            'uniform': [0, 1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['uniform'])

        # as string
        config_dict['uniform'] = '0:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1)

        # as dict
        config_dict['uniform'] = {'low': 0, 'high': 1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_quniform_option(self):
        def assert_equal(config, v1, v2, q, v3=None):
            result = {'low': v1, 'high': v2, 'q': q}
            if v3:
                result['size'] = v3
            assert config.to_dict()['quniform'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'quniform': [0, 1, 0.1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['quniform'])

        # as string
        config_dict['quniform'] = '0:1:0.1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1, 0.1)

        # as dict
        config_dict['quniform'] = {'low': 0, 'high': 1, 'q': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_loguniform_option(self):
        def assert_equal(config, v1, v2, v3=None):
            result = {'low': v1, 'high': v2}
            if v3:
                result['size'] = v3
            assert config.to_dict()['loguniform'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'loguniform': [0, 1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['loguniform'])

        # as string
        config_dict['loguniform'] = '0:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1)

        # as dict
        config_dict['loguniform'] = {'low': 0, 'high': 1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_qloguniform_option(self):
        def assert_equal(config, v1, v2, q, v3=None):
            result = {'low': v1, 'high': v2, 'q': q}
            if v3:
                result['size'] = v3
            assert config.to_dict()['qloguniform'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'qloguniform': [0, 1, 0.1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['qloguniform'])

        # as string
        config_dict['qloguniform'] = '0:1:0.1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1, 0.1)

        # as dict
        config_dict['qloguniform'] = {'low': 0, 'high': 1, 'q': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_normal_option(self):
        def assert_equal(config, v1, v2, v3=None):
            result = {'loc': v1, 'scale': v2}
            if v3:
                result['size'] = v3
            assert config.to_dict()['normal'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'normal': [0, 1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['normal'])

        # as string
        config_dict['normal'] = '0:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1)

        # as dict
        config_dict['normal'] = {'loc': 0, 'scale': 1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_qnormal_option(self):
        def assert_equal(config, v1, v2, q, v3=None):
            result = {'loc': v1, 'scale': v2, 'q': q}
            if v3:
                result['size'] = v3
            assert config.to_dict()['qnormal'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'qnormal': [0, 1, 0.1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['qnormal'])

        # as string
        config_dict['qnormal'] = '0:1:0.1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1, 0.1)

        # as dict
        config_dict['qnormal'] = {'loc': 0, 'scale': 1, 'q': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_lognormal_option(self):
        def assert_equal(config, v1, v2, v3=None):
            result = {'loc': v1, 'scale': v2}
            if v3:
                result['size'] = v3
            assert config.to_dict()['lognormal'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'lognormal': [0, 1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['lognormal'])

        # as string
        config_dict['lognormal'] = '0:1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1)

        # as dict
        config_dict['lognormal'] = {'loc': 0, 'scale': 1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict

    def test_matrix_qlognormal_option(self):
        def assert_equal(config, v1, v2, q, v3=None):
            result = {'loc': v1, 'scale': v2, 'q': q}
            if v3:
                result['size'] = v3
            assert config.to_dict()['qlognormal'] == result
            with self.assertRaises(ValidationError):
                config.to_numpy()
            with self.assertRaises(ValidationError):
                config.length
            assert isinstance(config.sample(), float)
            assert config.is_categorical is False
            assert config.is_distribution is True
            assert config.is_range is False
            assert config.is_uniform is False
            assert config.is_discrete is False
            assert config.is_continuous is True
            assert config.min is None
            assert config.max is None

        # as list
        config_dict = {
            'qlognormal': [0, 1, 0.1],
        }
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, *config_dict['qlognormal'])

        # as string
        config_dict['qlognormal'] = '0:1:0.1'
        config = MatrixConfig.from_dict(config_dict)
        assert_equal(config, 0, 1, 0.1)

        # as dict
        config_dict['qlognormal'] = {'loc': 0, 'scale': 1, 'q': 0.1}
        config = MatrixConfig.from_dict(config_dict)
        assert config.to_dict() == config_dict
